/*
 * jackalope -- universal control module
 * fnordnet communication protocol module
 *
 *   by Michael Schwab <ms@de-solution.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * For more information on the GPL, please go to:
 * http://www.gnu.org/copyleft/gpl.html
 */


/* fnord process -> wird immer aufgerufen */
/* empfangende frames werden nur verarbeitet wenn sie fertig empfangen wurden fnord_recv_complete == 1*/
/* extrahieren der daten aus dem empfangegen frame und einstellen in einen puffer */
/* auswertung ob ein frame erfolgreich uebertragen wurde -> ausloesen bestimmter aktionen*/

/*retransmit*/
/* wenn der fnord_recv_buf verarbeitet ist wird dort in den maintenance feldern ein flag gesetzt
erst dann kann ein neu empfangener frame in den recv buffer geschriebern werden
damit wird verhindert das ein frame der gerade verarbeiet wird berschrieben wird*/

#include <avr/pgmspace.h>
#include <stdio.h>
#include <avr/io.h>

#include "fnord-net.h"
#include "jackalope.h"
#include "rfm12.h"

#if FNORD_NET_ENABLE == 1

//#define FNORDRECV_TIMER0
//#define FNORDRECV_TIMER1

/* silence counter values for sending specified frame types */
/* in 10ms increment */
#define FNORD_SILENCE_COUNT_ACK 2
#define FNORD_SILENCE_COUNT_FRAME_MIN 15

// Frame segment lengh in bits
#define FNORD_FRAME_HEADER_LEN 24
#define FNORD_FRAME_CRC_LEN 16

/* define macros, if debug is enabled */
/* OLD SETUP
#if FNORD_NET_DEBUG == 1
    #define fnord_debug_printf(s, args...) fnord_debug_line++;if(fnord_debug_line>15){fnord_debug_line=0;};ansi_locate(fnord_debug_line+7,1);sprintf_P(uartbuf,PSTR(s), ## args);debug_output(uartbuf);
#else
    #define fnord_debug_printf(...)
#endif
*/

#if FNORD_NET_DEBUG == 1
    #define fnord_debug_printf(s, args...) sprintf_P(uartbuf,PSTR(s), ## args); serial_output(uartbuf);
#else
    #define fnord_debug_printf(...)
#endif


/* Sequenznummern verwaltung */
struct sequence_num_t
{
    union
    {
        uint8_t raw;
        struct
        {
            uint8_t tx:3;
            uint8_t txflag:1; // wird gesetzt wenn wir auf sein von uns gesendetes setup ein ack bekommen
            uint8_t rx:3;
            uint8_t rxflag:1;  // wird gesetzt wenn wir ein setup von dieser station erhalten haben
        };
    };
};


/* Fnord frame struktur */
struct fnord_frame_t
{
    union
    {
        uint8_t raw[21];
        struct
        {
            uint8_t protocol_version:3;
            uint8_t framelen:4; // data length 0 - 15 byte
            uint8_t setup:1;

            uint8_t ack:1;
            uint8_t to:6;
            uint8_t from:6;
            uint8_t sequence:3;

            /* max 15 byte data + 2 byte crc  */
            uint8_t data[17];
            /* fields wich are only used locally for frame management*/

            uint8_t ready_to_send:1;
            uint8_t ack_recv:1;
            uint8_t no_ack:1;
            uint8_t reserved:1;
            uint8_t retransmit:4;
        };
    };
};

typedef struct fnord_frame_t fnord_frame_t;
typedef struct sequence_num_t sequence_num_t;

static sequence_num_t seqnum[64];
static fnord_frame_t fnord_tx_buf;
/* receive buffer for complte frame*/
static fnord_frame_t fnord_recv_buf;
/* receive buffer for reception in progress*/
static fnord_frame_t fnord_ack_buf;

static uint8_t fnord_recv_buf_position;
static volatile uint8_t fnord_recv_complete;

static volatile uint8_t fnord_send_ping;
static volatile uint8_t fnord_silence_count;
static volatile uint8_t fnord_silence_count_frame = FNORD_SILENCE_COUNT_FRAME_MIN;

/* Flag is logical one when there is a frame which is currently being transmitted*/
static volatile uint8_t fnord_tx_in_progress;

/* Our local address 1 to 60 */
static uint8_t fnord_local_address;

//static uint8_t fnord_debug_line; // line position of debug output

static uint8_t fnord_tx_status;

// for values see function fnord_check_frame()
static uint8_t fnord_rx_status;

/* enduser receive buffer for fnordframe data */
static uint8_t fnord_rx_data[17]; // 1 Byte absender adresse + 16 Byte Data
static uint8_t fnord_rx_data_len; // anzahl bytes abzueglich der absender adresse

/* function prototypes */
static uint16_t crc_checksum(void *,uint8_t);
static uint8_t fnord_count_sequence(uint8_t);
static void recalc_fnord_silence_count_frame(void);
static uint8_t fnord_send_frame(fnord_frame_t *);
static uint8_t fnord_check_frame(fnord_frame_t *,uint8_t);

/* print fnordframe contents*/
void fnord_debug_frame(const char *,fnord_frame_t *);

/* create a fnord frame with data in it */
void fnord_compose_frame(fnord_frame_t *,uint8_t,uint8_t,uint8_t,uint8_t,uint8_t,uint8_t *,uint8_t);

/* CRC32 */
uint32_t crc32_update (uint32_t, uint8_t);

void fnord_init(uint8_t laddr)
{
    /* Fnord net sequence nummern initialisieren */
    for(uint8_t i=0;i<64;i++)
    {seqnum[i].raw=0;}

    fnord_local_address = laddr;
}

void fnord_debug_frame(const char *direction,fnord_frame_t *frame)
{
#if FNORD_NET_BASIC_DEBUG == 1 || FNORD_NET_DEBUG == 1
    strcpy_P(uartbuf,direction);
    serial_output(uartbuf);
    serial_output((char*)"\r\n");
#endif
#if FNORD_NET_DEBUG == 1
    fnord_debug_printf("DBG fnord frame Type:%u SETUP:%u ACK:%u SEQ:%u\r\n",frame->protocol_version, frame->setup, frame->ack,frame->sequence);
    fnord_debug_printf("DBG From:%u To:%u\r\n",frame->from, frame->to);
    fnord_debug_printf("DBG Datalen:%u CRC:0x%02x 0x%02x \r\n",
                  frame->framelen, frame->data[frame->framelen], frame->data[frame->framelen+1]);
#endif
}

// this function takes care of all receive and transmit and decoding actions
// call it at least every 200ms
void fnord_process(void)
{
    fnord_tx_status = 0;

    if (fnord_recv_complete == 1)
    {
        fnord_recv_complete = 0;

        fnord_debug_frame(PSTR("RX Frame:"),&fnord_recv_buf);

        fnord_rx_status = fnord_check_frame(&fnord_recv_buf,fnord_recv_buf_position);
        fnord_debug_printf("RX FnordFrame Check Status: %u\r\n\r\n",fnord_rx_status);

/*
        if ( fnord_rx_status != 40) // Status == 40 wenn Frame NICHT an unsere adresse
        {
            fnord_debug_printf("RX Fnord Frame Type: %u Len: %u\r\n",fnord_recv_buf.protocol_version,fnord_recv_buf.framelen);
            fnord_debug_printf("SETUP:%u ACK:%u SEQ:%u FROM:%u TO:%u BITS:%u\r\n",
                    fnord_recv_buf.setup,
                    fnord_recv_buf.ack,
                    fnord_recv_buf.sequence,
                    fnord_recv_buf.from,
                    fnord_recv_buf.to,
                    fnord_recv_buf_position);

            //fnord_debug_printf("Frame Check Status:%u\r\n",fnord_rx_status);
        }
*/

        if ( fnord_rx_status == 0 )
        {
            /* setup frame received */
            if (fnord_recv_buf.setup == 1)
            {
                /* got ACK for a setup frame we send */
                if (fnord_recv_buf.ack == 1)
                {
                    /* Initialize sequence numbers for this destination */
                    seqnum[fnord_recv_buf.from].tx = 0;
                    seqnum[fnord_recv_buf.from].txflag = 1;
                    fnord_tx_buf.ack_recv = 1 ;
                    fnord_tx_status = 1;
                    //fnord_debug_printf("RX SETUP ACK FROM:%u\r\n",fnord_recv_buf.from);
                }
                else
                {
                    /* received a SETUP request */
                    /* answer with an ACK */
                    fnord_compose_frame(&fnord_ack_buf,1,1,1,fnord_local_address,fnord_recv_buf.from,(uint8_t *)"",0);
                    fnord_ack_buf.ready_to_send=1;
                    fnord_ack_buf.retransmit=1 ;

                    /* set rxflag for this destination */
                    seqnum[fnord_recv_buf.from].rx = 0;
                    seqnum[fnord_recv_buf.from].rxflag = 1;

                    //fnord_debug_printf("FNORD TX SETUP ACK TO:%u\r\n",fnord_ack_buf.to);
                }
            }
            else
            {
                /* Frames mit SETUP = 0 = normale daten frames */
                /* wenn ein ACK mit der richtigen sequenznummer empfangen wurde, dann auswerten */
                if (fnord_recv_buf.ack == 1)
                {
                  /* this is an ack for a frame we have send */
                  /* Set ack flag in frame maintenace field for later processing*/
                  fnord_tx_buf.ack_recv = 1;
                  /* Increase sequence number to use for sending next frame to this destination */
                  seqnum[fnord_tx_buf.to].tx = fnord_count_sequence(seqnum[fnord_tx_buf.to].tx);
                  fnord_debug_printf("Fnord RX ACK FROM:%u  SEQ:%u\r\n",fnord_recv_buf.from,fnord_recv_buf.sequence);
                  fnord_tx_status = 1; // Tx of data frame OK
                }
                else
                {
                  /* send ACK for this frame */
                  fnord_compose_frame(&fnord_ack_buf,1,0,1,fnord_local_address,fnord_recv_buf.from,(uint8_t *)"",0);
                  /* set correct seq num */
                  fnord_ack_buf.sequence = fnord_recv_buf.sequence;
                  /* set flag that transmission takes place after processing received frame */
                  fnord_ack_buf.ready_to_send=1 ;
                  fnord_ack_buf.retransmit=1 ;

                  // FIXME Empfangene Daten nur auswerten wenn diese noch nicht empfangen
                  // wurde das kann anhand der sequenznummer geprft werden
                  if ( seqnum[fnord_recv_buf.from].rx == fnord_recv_buf.sequence)
                  {
                        // Diese sequenznummer haben wir noch nicht empfangen
                        // daten auswerten
                        fnord_rx_data_len = fnord_recv_buf.framelen;

                        /* Increase sequence number we expect in next frame from this source */
                        // Sequenznummer nur weiterzaehlen wenn wir einen frame erhalten haben der neu war
                        // und verarbeitet wurde, wenn dieser frame nochmal empfangen wird weil z.B. unser ack
                        // verloren gegangen ist dann wird ein ack gesendet es findet aber keine erneute
                        // auswertung und verarbeitung statt
                        seqnum[fnord_recv_buf.from].rx = fnord_count_sequence(fnord_recv_buf.sequence);
                        fnord_debug_printf("FNORD TX ACK TO:%u SEQ: %u\r\n",fnord_ack_buf.to,fnord_ack_buf.sequence);
                  }
                  else
                  {
                      // das ist ein retransmit daten nicht auswerten
                      fnord_rx_data_len = 0;
                  }

                }

                /* wenn daten empfangen wurden dann speichern */
                if (fnord_rx_data_len > 0)
                {
                    for (uint8_t i=0; i<fnord_rx_data_len; i++)
                    {
                        #if FNORD_NET_DEBUG_DATA == 1
                        fnord_debug_printf("FNORD DATA: BYTE: %u VALUE: %u\r\n",i,fnord_recv_buf.data[i]);
                        #endif
                        fnord_rx_data[i+1] = fnord_recv_buf.data[i] ;
                    }
                    // fist byte of data is the address of the sender
                    fnord_rx_data[0] = fnord_recv_buf.from;
                    // Terminate data with NULL byte so that it can be processed as string
                    fnord_rx_data[fnord_rx_data_len+1] = 0x00;
                }
            } /* ende if frames mit SETUP = 0 */

        } /* ende if framecheck is OK */

        fnord_recv_buf_position = 0;
    } // ende if fnordframe empfangen => fnord_recv_complete == 1

     /* wenn der silence counter hoch genug steht kann genau ein Frame gesendet werden
        ack frames werden bevorzugt
        der counter wird durch jeden Receive TIMER Overflow Interrupt eins hochgezaehlt
        ein TIMER overfolw findet alle 820us statt */
    if (fnord_silence_count >= FNORD_SILENCE_COUNT_ACK && fnord_tx_in_progress == 0)
    {
        //debug_printf("Fnord SILENCE COUNT:%u\r\n",fnord_silence_count);
        /* ACK Frame senden wenn noetig */
        if (fnord_ack_buf.ready_to_send == 1 && fnord_ack_buf.retransmit > 0 && fnord_check_collision() == 0)
        {
            fnord_send_frame(&fnord_ack_buf);
            fnord_ack_buf.retransmit = 0 ;
        }
        else
        {
            /* wenn kein ack frame gesendet wurde dann kann ein fnord daten frame gesendet werden*/
            /* Fnord frame aus tx buffer senden wenn einer da ist */
            if (fnord_tx_buf.ready_to_send == 1 &&
                fnord_tx_buf.ack_recv == 0 &&
                fnord_tx_buf.retransmit > 0 &&
                fnord_silence_count >= fnord_silence_count_frame &&
                fnord_check_collision() == 0)
            {
              fnord_send_frame(&fnord_tx_buf);
              fnord_tx_buf.retransmit-- ;
              recalc_fnord_silence_count_frame();
            }
        } // ende if kein ack frame gesendet
    } // ende if fnord_silence_count

    // Uebertragung des frames der sich in fnord_tx_buf befindet fehlgeschlagen
    if (fnord_tx_buf.retransmit == 0 &&
        fnord_tx_buf.ready_to_send == 1 &&
        fnord_tx_buf.ack_recv == 0 &&
        fnord_tx_buf.no_ack == 0 &&
        fnord_silence_count > fnord_silence_count_frame)
    {
        fnord_debug_printf("No ACK -> TARGET NOT RESPONDING\r\n");
        /* out_printf("ERROR: dead target, select again\r\n"); */
        fnord_tx_buf.no_ack = 1;
        /* Ziel muss neu initialisiert werden */
        seqnum[fnord_tx_buf.to].tx = 0;
        seqnum[fnord_tx_buf.to].txflag = 0;
        fnord_tx_status = 10;
    }
} /* ende fnord process */


/* fnordframe zusammenbauen */
void fnord_compose_frame(fnord_frame_t *frame, uint8_t protocol_version, uint8_t setup, uint8_t ack, uint8_t from,uint8_t to, uint8_t *data, uint8_t datalen)
{

    frame->protocol_version = protocol_version;
    frame->setup = setup;
    frame->ack  = ack;
    frame->framelen = datalen;
    frame->to   = to;
    frame->from = from;

    if (ack == 1) // ACK frame nur einmal senden
    {frame->retransmit = 1;}
    else // datenframe bis zu 4 mal senden
    {frame->retransmit = 4;}

    // The sequence number will be set directly before the frame is send
    // in function fnord_send_frame
    // frame->sequence = seqnum[to].tx;

    for(uint8_t i=0;i<datalen;i++)
    { frame->data[i] = data[i]; }

    frame->ready_to_send = 1;
    frame->ack_recv = 0;
    frame->no_ack = 0 ;
}

void fnord_transmit_frame(uint8_t send_to, uint8_t *payload, uint8_t payloadlength)
{
    if (send_to > 0 && send_to < 64)
    {
        fnord_compose_frame(&fnord_tx_buf,1,0,0,fnord_local_address,send_to,payload,payloadlength);
    }
}

void fnord_transmit_setup(uint8_t send_to)
{
    if (send_to > 0 && send_to < 64)
    {
        fnord_compose_frame(&fnord_tx_buf,1,1,0,fnord_local_address,send_to,(uint8_t *)"",0);
    }
}

uint8_t fnord_check_setup(uint8_t destination)
{
    if (destination > 0 && destination < 64)
    {
        return seqnum[destination].txflag;
    }
    else
    {
        return 2;
    }
}

/* Fnordframe verschicken */
/* lowlevel for sending a frame */
uint8_t fnord_send_frame(fnord_frame_t *frame)
{
    uint16_t checksum;
    uint8_t bitcount;

    fnord_tx_in_progress = 1;

    bitcount = (frame->framelen<<3)  + FNORD_FRAME_HEADER_LEN + FNORD_FRAME_CRC_LEN; // datelen * 8 + 3 byte header + 2 byte crc

    /* Set serial number only for non ACK and non SETUP frames*/
    if (frame->ack == 0 && frame->setup == 0)
    {
        frame->sequence = seqnum[frame->to].tx;
    }
    else if (frame->setup == 1)
    {
        // setup frames do always have sequence 0
        frame->sequence = 0;
    }
    else if (frame->ack == 1 && frame->setup == 0)
    {
        // fnord_process() takes care of that
        //frame->sequence = fnord_recv_buf.sequence;
    }

    /* do not send setup = 1 = setup frame with data*/
    if ( frame->setup == 1 && frame->framelen > 0)
    {return 128;}

    /* calculate checksum */
    checksum = crc_checksum(frame, 3 + frame->framelen);

    /* put checksum behind data */
    frame->data[frame->framelen] = LO8(checksum);
    frame->data[frame->framelen+1] = HI8(checksum);


    fnord_debug_frame(PSTR("TX Frame:"),frame);

/*
    fnord_debug_printf("TX fnord frame Type:%u SETUP:%u ACK:%u SEQ:%u\r\n",frame->protocol_version, frame->setup, frame->ack,frame->sequence);
    fnord_debug_printf("TX From:%u To:%u\r\n",frame->from, frame->to);
    fnord_debug_printf("TX Datalen:%u CRC:0x%02x 0x%02x BITCNT: %u\r\n",
                  frame->framelen, frame->data[frame->framelen], frame->data[frame->framelen+1],bitcount);
*/

    #if FNORD_NET_DEBUG_DATA == 1
    fnord_debug_printf("sending fnord frame bitpattern: \r\n");
    for (uint8_t i = 0; i < bitcount; i++)
    {
        uint8_t bit, byte;
        byte = i / 8;
        bit = i % 8;
        fnord_debug_printf("|%u|",frame->raw[byte] & _BV(bit));
        if (bit == 7)
        { fnord_debug_printf("\r\n");}
    }
    #endif

    fnord_debug_printf("\r\n");


    //ir_send_bitstream(frame,bitcount);
    //rfm12_tx_bytestream(frame->raw,(bitcount/8));

    // wenn fonrd-net einen frame senden mchte muss es an
    // das Haputprogramm einen Pointer auf einen Array bergeben der dann
    // vom Hauptprogramm an rfm12 oder sonstwohin weitergereicht wird
    // dazu ruft fnord-net folgende Funktion auf
    // void fnord_tx_request(uint8_t *,datalen);

    if (frame->ack == 0)
    { fnord_tx_status = 0; }

    fnord_tx_request(&frame->raw[0],(bitcount/8));

    return 0;
}


/* check if fnord frame is valid */
/* return 0 on ok and errorcode on fail */
uint8_t fnord_check_frame(fnord_frame_t *frame,uint8_t bitcount)
{
  //fnord_debug_printf("Fnord Check Frame\r\n");
  if ((bitcount % 8 )!= 0)
  {
    fnord_debug_printf("Fnord frame invalid bitcount: %u\r\n",bitcount);
    return 10;
  }

    if (frame->framelen * 8 + FNORD_FRAME_HEADER_LEN + FNORD_FRAME_CRC_LEN != bitcount)
    {
        fnord_debug_printf("Fnord frame len vs. bitcount mismatch len: %u\r\n",frame->framelen);
        return 20;
    }

  uint16_t checksum = crc_checksum(frame, (bitcount/8)-2);

  if ( LO8(checksum) != frame->raw[(bitcount/8)-2] || HI8(checksum) != frame->raw[(bitcount/8)-1] )
  {
    fnord_debug_printf("Fnord frame CRC error\r\n");
    return 30;
  }

  /* Found a physically valid frame now check the contents for validity */

  if (frame->to != fnord_local_address)
  {
    //fnord_debug_printf("Fnord frame not for us -> DROPPED\r\n");
    return 40 ;
  }

  /* check for valid sequence number */
  if (frame->setup == 1)
  {
    /* Check setup frame */

    if ( fnord_tx_buf.setup == 0 && frame->ack == 1)
    {
     fnord_debug_printf("RX SETUP ACK from %z without TX SETUP Frame\r\n",frame->from);
     return 51;
    }

    /* invalid parameters in setup frame */
    if ( frame->sequence != 0 || frame->framelen > 0)
    {
     fnord_debug_printf("RX invalid SETUP Frame from %z seq !=0 or datalen !=0\r\n",frame->from);
     return 50;
    }
  }
  else
  {

    /* check data frame with ACK Flag set*/
    if (frame->ack == 1)
    {
        /* received a frame with ACK bit set which is not a setup frame */
        /* this frame shoud have the seq number which was used in the originating frame */
        if (frame->sequence != fnord_tx_buf.sequence )
        {
            fnord_debug_printf("Fnord ACK Frame Sequence error IS:%u SHOULD BE:%u\r\n",frame->sequence,fnord_tx_buf.sequence);
            return 70;
        }

        if (frame->from != fnord_tx_buf.to )
        {
            /*
            fnord_debug_printf("Fnord ACK Frame Address error IS:%u SHOULD BE:%u\r\n",frame->from,fnord_tx_buf.to);
            fnord_debug_printf("TX BUF fnord frame Type:%u SETUP:%u ACK:%u SEQ:%u\r\n",fnord_tx_buf.protocol_version, fnord_tx_buf.setup, fnord_tx_buf.ack,fnord_tx_buf.sequence);
            fnord_debug_printf("TX BUF From:%u To:%u\r\n",fnord_tx_buf.from, fnord_tx_buf.to);
            fnord_debug_printf("TX BUF Datalen:%u CRC:0x%02x 0x%02x\r\n",fnord_tx_buf.framelen, fnord_tx_buf.data[fnord_tx_buf.framelen], fnord_tx_buf.data[fnord_tx_buf.framelen+1]);
            */
            return 75;
        }

    }
    else
    {
        /* Check data frame */
        if ( seqnum[frame->from].rxflag == 0 )
        {
            // Destination not initialized
            fnord_debug_printf("Dataframe without previous setup received\r\n");
            return 60;
        }

        /* Frame seqnum = stored seqnum or stored seqnum +1 when the frame is a retransmit */
        if (frame->sequence != seqnum[frame->from].rx &&
            fnord_count_sequence(frame->sequence) != seqnum[frame->from].rx)
        {
            fnord_debug_printf("Fnord Data Frame Sequence error IS:%u SHOULD BE:%u\r\n",frame->sequence,seqnum[frame->from].rx);
            return 65;
        }
    } // ende else if frame->ack == 1
   } // ende else if frame->setup == 1
 return 0;
}

uint8_t fnord_count_sequence(uint8_t seq)
{
      if (seq == 7)
      { return 0; }
      else
      { seq++;}

  return seq;
}

void recalc_fnord_silence_count_frame(void)
{
    fnord_silence_count_frame = FNORD_SILENCE_COUNT_FRAME_MIN + ( rand() % 30);
    fnord_debug_printf("New silence count %u\r\n",fnord_silence_count_frame);
}

uint16_t crc_checksum(void *data, uint8_t length)
{
    uint16_t crc = 0;
    uint8_t *p = (uint8_t *)data;

    for (uint8_t i = 0; i < length; i++) {
        //crc = _crc_ibutton_update(crc, *p);
        //crc = _crc_ccitt_update(crc, *p);
        crc = _crc16_update(crc, *p);
        p++;
    }
    return crc;
}

void fnord_countup_silence_count(void)
{
    fnord_silence_count++;
    if ( fnord_silence_count == 0xff) { fnord_silence_count = 0xfe; }

}

void fnord_clear_silence_count(void)
{
    fnord_silence_count = 0;
}

void fnord_set_local_address(uint8_t addr)
{
    fnord_local_address = addr;
}

/* status of frame to be transmitted,0=undef, 1=TX OK, 10=TX fail */
uint8_t fnord_get_tx_status(void)
{
    return fnord_tx_status;
}

void set_fnord_tx_in_progress(uint8_t value)
{
    fnord_tx_in_progress = value;
}

// for values see function fnord_check_frame()
uint8_t fnord_get_rx_status(void)
{
    return fnord_rx_status;
}

 // clears rx_datalen returns a pointer to rx_data array
uint8_t fnord_get_rx_datalen(void)
{
    uint8_t datalen;
    datalen = fnord_rx_data_len;
    fnord_rx_data_len = 0;
    return datalen;
}

/* enduser receive for fnordframe data */
uint8_t *fnord_get_rx_data(void)
{
    return fnord_rx_data;
}

void fnord_received_frame(uint8_t *data,uint8_t length)
{
    fnord_debug_printf("RFM FRAMELEN: %u\r\n",length);
    // Laenge auf framelaenge+1 (= erstes byte = laenge) limitieren
    if ( length > ( sizeof(fnord_recv_buf.raw)+1) )
    {
        length = (sizeof(fnord_recv_buf.raw)+1);
    }

    // Erstes byte ueberspringen da das die anzahl der frame bytes enthaelt
    for ( uint8_t i =1; i < length; i++)
    {
        fnord_recv_buf.raw[i-1] = data[i];
        //debug_printf("FNORD RAW DATA: %d val: %d\r\n",i,data[i]);
    }
    // Fake correct Framelength
    fnord_recv_buf_position = (fnord_recv_buf.framelen * 8) + FNORD_FRAME_HEADER_LEN + FNORD_FRAME_CRC_LEN;
    fnord_recv_complete = 1;
}
#endif

// https://www.lammertbies.nl/comm/info/crc-calculation.html
// TODO Test der routine funzt eventuell nicht
uint32_t crc32_update (uint32_t crc, uint8_t byte)
{
    uint8_t i;
    crc = crc ^ byte;
    for (i = 0; i < 8; i++)
    {
        uint8_t bit0 = crc & 0x01;
        crc = (crc >> 1);
        if (bit0 == 1)
        {
            crc ^= 0xedb88320ul;
        }
    }
    return crc;

    /*
    uint8_t i;
    uint8_t crc_neu;
    crc = crc ^ byte;
    for (i = 0; i < 8; i++)
    {
        crc_neu = (crc >> 1);
        if ( ( crc & 0x01) != 0)
        {
            crc_neu ^= 0xedb88320ul;
        }
        crc = crc_neu;
    }
    return crc;
    */

    /*
    int8_t i;
    crc = crc ^ byte;
    for(i=7; i>=0; i--)
    crc=(crc>>1)^(0xedb88320ul & (-(crc&1)));
    return(crc);
    */
}

uint32_t crc32_checksum(void *data, uint8_t length)
{
    uint32_t crc = 0xffffffff;
    uint8_t *p = (uint8_t *)data;

    for (uint8_t i = 0; i < length; i++) {
        //debug_printf("POS: %u DATA: %02x \r\n",i,*p);
        crc = crc32_update(crc, *p);
        p++;
    }
    crc = crc^0xffffffff;
    return crc;
}
