/*
 * jackalope -- universal control module
 *
 *   by Michael Schwab <ms@de-solution.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * For more information on the GPL, please go to:
 * http://www.gnu.org/copyleft/gpl.html
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <stdlib.h>
#include <stdio.h>
#include <util/crc16.h>
#include <string.h>

#include "jackalope.h"
#include "eeprom.h"
#include "dht11.h"
//#include "onewire/onewire.h"

#if RFM12_ENABLE == 1
    #include "rfm12.h"
#endif

#if RFM69_ENABLE == 1
    #include "rfm69.h"
#endif

#if FNORD_NET_ENABLE == 1
    #include "fnord-net.h"
#endif

/* prototypes */
// Bootloder Start Addreess
// compiled into application so that the application can jump into bootloader
#define BOOTLOADER_START 0x7c00

void (*jump_to_bootloader)(void) = (void *)BOOTLOADER_START;

/* fnord_config.struktur die im eeprom abgelegt wird*/
struct fnord_config_t
{
    union
    {
        uint8_t raw[14];
        struct
        {
            uint8_t  local_address;
            uint8_t  rssi;
            uint16_t baudrate;
            uint8_t  channel;
            uint8_t  txpower;
            uint8_t  adcref;
            uint8_t  json;
            uint32_t its_addr;
            uint8_t  rfmzin;
            uint8_t  debuglevel;
        };
    };
};

struct fnord_config_t fnord_config;

/*
 * Struktur die zum Parsen eingegebener Kommandos und deren Ausfuehrung
 * herangezogen wird
 */
struct cli_cmd_t
{
    PGM_P cmd_text; // Befehlstext
    PGM_P return_text; // Return Text des kommandos
    PGM_P error_text;  // Error Text des kommandos
#if HELP_ENABLE == 1
    PGM_P help_text;  // Hilfe Text des kommandos
#endif
    uint8_t parametercount; // Anzahl der Parameter die uebergeben werden muessen
    uint8_t always_local; // Dieser Befehl wirkt immer auf die lokale hardware
    void (*func)(struct cmd_data_t *); // Function pointer auf die Funktion die das kommando implementiert
};

struct cli_cmd_t cli_cmd[50];

/* Variablen die duch getmem / setmem beinflusst werden knnen
 * zur Verwendung in programmen die direkt hier laufen funktion exec_internal_programm
 * mit merker.wert[0] lsst sich einstellen ob das interne programm ausgefhrt wird oder nicht*/
struct merker_data_t
{
    union
    {
        uint8_t raw[22];
        struct
        {
            int16_t wert[11];
        };
    };
};

struct merker_data_t merker;

#if RFM69_OOK == 1
volatile uint8_t ti1_ovf_cnt = 0;
volatile uint8_t ook_capture = 0;
volatile uint16_t ooktime_count;
volatile uint8_t pinlevel_pb0;
volatile uint8_t ook_repeat_count = 0;
uint32_t ook_message;
uint8_t ook_protocol;
#endif

/* Array mit moeglichen Fehlermeldungen */
PGM_P errormessage[15];

/* Prototypen */
static void init(void);
static void adcinit(void);
static void spi_init(void);
static void set_cfg_defaults(void);
void display_binary(uint8_t);

#if FNORD_NET_ENABLE == 1
static uint8_t fnordnet_tx_status(void);
#endif

static void cli_cmd_init(void);
static void cli_err_init(void);

/* command functions */
static void cmd_version(struct cmd_data_t *cmd);
static void cmd_test(struct cmd_data_t *cmd);
static void cmd_crc32(struct cmd_data_t *cmd);
static void cmd_savecfg(struct cmd_data_t *cmd);
static void cmd_loadcfg(struct cmd_data_t *cmd);
static void cmd_savemem(struct cmd_data_t *cmd);
static void cmd_out(struct cmd_data_t *cmd);
static void cmd_in(struct cmd_data_t *cmd);
static void cmd_adcin(struct cmd_data_t *cmd);
static void cmd_adcinall(struct cmd_data_t *cmd);
static void cmd_adcavgin(struct cmd_data_t *cmd);
static void cmd_adcavginall(struct cmd_data_t *cmd);
static void cmd_adcsetref(struct cmd_data_t *cmd);
static void cmd_dacout(struct cmd_data_t *cmd);
static void cmd_getmem(struct cmd_data_t *cmd);
static void cmd_setmem(struct cmd_data_t *cmd);
static void cmd_memfree(struct cmd_data_t *cmd);

static void cmd_adclist(struct cmd_data_t *cmd);
static void cmd_help(struct cmd_data_t *cmd);

static void cmd_bootloader(struct cmd_data_t *cmd);

#if DHT11_ENABLE == 1
static void cmd_temp(struct cmd_data_t *cmd);
#endif

#if RFM69_DEBUG_CMDS == 1 && RFM69_ENABLE == 1
static void cmd_rfm(struct cmd_data_t *cmd);
static void cmd_rfm_write(struct cmd_data_t *cmd);
static void cmd_fifow(struct cmd_data_t *cmd);
static void cmd_fifor(struct cmd_data_t *cmd);
#endif

#if RFM69_OOK == 1
static void cmd_toom(struct cmd_data_t *cmd);
static void cmd_ooksend(struct cmd_data_t *cmd);
static void cmd_itsset(struct cmd_data_t *cmd);
void ook_sendbit(uint8_t, uint8_t);

#if RFM69_OOK_CAPTURE == 1
static void cmd_ookcapture(struct cmd_data_t *cmd);
#endif

void rfm69_diable_capture(void);
void addbit(uint8_t *,uint16_t *,uint8_t);
void ook_send_message(void);
#if RFM69_OOKMODE == 1
void ook_sendbit(uint8_t, uint8_t);
#endif
#endif

/* http://www.eckart-winkler.de/computer/c/varargs.htm */
int16_t exec_cmd(const char *, int16_t , ... );

#if RFM69_ENABLE == 1
static void cmd_frequency(struct cmd_data_t *cmd);
static void cmd_rfmtemp(struct cmd_data_t *cmd);
static void cmd_rfmzin(struct cmd_data_t *cmd);
#endif

#if RFM12_ENABLE == 1 || RFM69_ENABLE == 1
static void cmd_defaults(struct cmd_data_t *cmd);
static void cmd_rfms(struct cmd_data_t *cmd);
static void cmd_channel(struct cmd_data_t *cmd);
static void cmd_baudrate(struct cmd_data_t *cmd);
static void cmd_txpower(struct cmd_data_t *cmd);
static void cmd_rssi(struct cmd_data_t *cmd);
static void cmd_rssi_scan(struct cmd_data_t *cmd);
static void cmd_rssi_survey(struct cmd_data_t *cmd);
static void cmd_setmod(struct cmd_data_t *cmd);
#else
// dummy function for transmitting a fnord frame which will never happen
void fnord_tx_request(uint8_t *,uint8_t);
#endif

#if FNORD_NET_ENABLE == 1
static void cmd_sla(struct cmd_data_t *cmd);
static void cmd_setup(struct cmd_data_t *cmd);
static void cmd_select(struct cmd_data_t *cmd);
static void cmd_status(struct cmd_data_t *cmd);
#endif

/* Variablen */
static volatile uint8_t systemcounter = 0;
static volatile uint8_t ui_display_flag = 0;

/* Variablen fuers CLI */
char command[45];

uint16_t adcval;
uint8_t fnord_remote_address; // adresse der angewaehlten Gegenstation wenn fnordnet aktiv ist
uint8_t ui_cmd_ready;
uint8_t ui_uart_rx_pointer;
uint8_t ui_uart_rx_pointer_old;

int main(void)
{
    uart_init();
    spi_init();
    init();
    cli_cmd_init();
    cli_err_init();
    sei();

    uint8_t i;

    out_printf("Jackalope starting...\r\n");


#if RFM69_ENABLE == 1
    // Check for RFM69 RF Module
    i = rfm69_get_chiprevision();
    out_printf("RFM69CW chip revision %u",i);

    if (i == 0 || i == 15)
    {
        out_printf(" invalid version RFM69CW possibly not connected",i);
    }
    else
    {
        out_printf(" OK");
    }
    out_printf("\r\n");
#endif


// daten aus dem eeprom lesen wenn das crc nicht stimmt dann defaults verwenden

    uint16_t status;
    status = eeprom_read_bytearray(fnord_config.raw,sizeof(fnord_config.raw),0);
    if (status != sizeof(fnord_config.raw))
    {
        out_printf("Setting fnordnet default values\r\n");
        set_cfg_defaults();
    }
    else
    {
#if RFM12_ENABLE == 1 || RFM69_ENABLE == 1
        out_printf("Loading fnordnet EEPROM values\r\n");
        cmd_data.value[0] = (uint8_t)fnord_config.txpower;
        cmd_txpower(&cmd_data);
        cmd_data.value[0] = (uint8_t)fnord_config.channel;
        cmd_channel(&cmd_data);
#endif
    }

    /* ADC default values requires fnord_config to be valid because of Vref selection */
    adcinit();

    status = eeprom_read_bytearray(merker.raw,sizeof(merker.raw),sizeof(fnord_config.raw)+2);
    if (status != sizeof(merker.raw))
    {
        out_printf("Setting mem default values\r\n");
        // Merker initialisieren
        for (i = 0; i < 11; i++)
        {
            merker.wert[i] = 0;
        }
    }
    else
    {
        out_printf("Loading mem EEPROM values\r\n");
    }

    debug_printf("Debug enabled\r\n");

#if FNORD_NET_ENABLE == 1
        fnord_set_local_address(fnord_config.local_address);
        srand(fnord_config.local_address);
#endif

#if RFM12_ENABLE == 1
        rfm12_setrssi(fnord_config.rssi);
        rfm12_setbaudrate(fnord_config.baudrate);
        rfm12_setchannel(fnord_config.channel);
        rfm12_settxpower(fnord_config.txpower);
        rfm12_init();
        debug_printf("RFM12: ADDR: %d RSSI: %d BAUD: %d CHAN %d TX: %d\r\n",fnord_config.local_address,fnord_config.rssi,fnord_config.baudrate,fnord_config.channel,fnord_config.txpower);
#endif

#if RFM69_ENABLE == 1
        // Wird durch cmd_loadcfg(&cmd_data) weiter unten erledigt
        //rfm69_settxpower(fnord_config.txpower);
        //rfm69_setchannel(fnord_config.channel);
        //rfm69_setrssi(fnord_config.rssi);
        //rfm69_setbaudrate(fnord_config.baudrate);
        rfm69_init();
        debug_printf("RFM69: ADDR: %d RSSI: %d BAUD: %d CHAN %d TX: %d\r\n",fnord_config.local_address,fnord_config.rssi,fnord_config.baudrate,fnord_config.channel,fnord_config.txpower);
#endif

    cmd_data.value[0] = 0;
    cmd_data.value[1] = 0;
    cmd_data.value[2] = 0;
    cmd_data.value[3] = 0;
    cmd_loadcfg(&cmd_data);
    out_printf("RFM: ADDR: %d RSSI: %d BAUD: %d CHAN %d TX: %d\r\n",fnord_config.local_address,fnord_config.rssi,fnord_config.baudrate,fnord_config.channel,fnord_config.txpower);

    // in adcinit wird die Vref gesetzt
    out_printf("Setting ADC Ref %d Volt\r\n",fnord_config.adcref);
    out_printf("Setting RFM Zin %d Ohm\r\n",fnord_config.rfmzin);
    out_printf("Setting JSON Output %d\r\n",fnord_config.json);
    out_printf("Jackalope version %d.%d ready\r\n",MAJOR_VERSION,MINOR_VERSION);

    //onewire_init();

    while (1)
    {
        // Kommando auswerten

        if ( ui_cmd_ready == 1)
        {
            process_cli_command();
        }

        if ( ui_display_flag == 1)
        {
            ui_display();
            ui_display_flag = 0;
        }

        /* Programm ausfuehren das einkompiliert wurde */
        /* ber kommandozeile merker 0 (setmem 0 1) einschaltbar */
        if (merker.wert[0] == 1)
        {
            exec_internal_programm();
        }

#if FNORD_NET_ENABLE == 1
        fnord_process();
        if ( fnord_get_tx_status() == 10 )
        {
            if ( fnord_config.json == 1 )
            {
                out_printf("{\"error\":\"3\", \"errortext\":\"ERROR: Remote site does not answer\"}\r\n");
            }
            else
            {
                out_printf("ERROR: Remote site does not answer\r\n");
            }
        }

        // when we received fnord frame
        // process received command
        uint8_t datalen = fnord_get_rx_datalen();
        if (datalen == 11)
        {
            // data = Byte0 is remote address and 1...n frame paylod data
            uint8_t *data = fnord_get_rx_data();
            for (i=0; i<12; i++)
            {
                cmd_data.raw[i] = data[i];
            }

            // Antwort auf ein gesedetes Kommando
            if (cmd_data.cmd_id == 255)
            {
#ifdef DEBUG
                debug_printf("CMD RESULT From: %d CMD: ",cmd_data.remote_addr);
                strcpy_P(uartbuf,cli_cmd[cmd_data.sub_command].cmd_text);
                serial_output(uartbuf);
                debug_printf(" Data: %d %d %d %d Err: %d\r\n",cmd_data.value[0], cmd_data.value[1], cmd_data.value[2], cmd_data.value[3], cmd_data.err);
#endif
                if (fnord_config.json == 0)
                {

                    if ( cmd_data.err < 2)
                    {
                        sprintf_P(uartbuf,cli_cmd[cmd_data.sub_command].return_text,cmd_data.value[0],cmd_data.value[1],cmd_data.value[2],cmd_data.value[3]);
                        serial_output(uartbuf);
                        serial_output((char*)"\r\n");
                    }
                    else
                    {
                        if ( cmd_data.err > 14 ) {cmd_data.err = 14;}
                        sprintf_P(uartbuf,errormessage[cmd_data.err],cmd_data.value[0],cmd_data.value[1],cmd_data.value[2],cmd_data.value[3]);
                        serial_output(uartbuf);
                        serial_output((char*)"\r\n");
                    }
                }
                else if (fnord_config.json == 1)
                {
                    if ( cmd_data.err < 2)
                    {
                        sprintf_P(uartbuf,PSTR("{ \"value1\":\"%d\",\"value2\":\"%d\",\"value3\":\"%d\",\"value4\":\"%d\" }"),cmd_data.value[0],cmd_data.value[1],cmd_data.value[2],cmd_data.value[3]);
                        serial_output(uartbuf);
                        serial_output((char*)"\r\n");
                    }
                    else
                    {
                        if ( cmd_data.err > 14 ) {cmd_data.err = 14;}
                        sprintf_P(uartbuf,PSTR("{ \"error\":\"%u\", \"errortext\":\"%s\" }"),cmd_data.err,errormessage[cmd_data.err],cmd_data.value[0],cmd_data.value[1],cmd_data.value[2],cmd_data.value[3]);
                        serial_output(uartbuf);
                        serial_output((char*)"\r\n");
                    }
                }
                else
                {
                    out_printf("JSON / TXT output cfg not set\r\n");
                }
            }
            else if (cmd_data.cmd_id == 254)
            {
                // Daten die uns ohne Aufforderung gesendet wurden
                // dass kann sein wenn ein Jackalope periodisch Statusdaten
                // an eine Zentrale sendet, die dort ausgewertet werden
                // MUSS NOCH IMPLEMENTIERT WERDEN
            }
            else if (cli_cmd[cmd_data.cmd_id].parametercount > 0)
            {
#ifdef DEBUG
                debug_printf("RCV CMD From: %d CMD: ",cmd_data.remote_addr);
                strcpy_P(uartbuf,cli_cmd[cmd_data.cmd_id].cmd_text);
                serial_output(uartbuf);
                debug_printf(" Data: %d %d %d %d Err: %d\r\n",cmd_data.value[0], cmd_data.value[1], cmd_data.value[2], cmd_data.value[3], cmd_data.err);
#endif
                // Kommando von remote an uns

                // Als erstes checken ob der Absender erreichbar ist
                uint8_t f_tx_status;
                if ( fnord_check_setup(cmd_data.remote_addr) == 0 )
                {
                    // wir haben noch nie mit dem Absender gesprochen also setup senden
                    debug_printf("Target %d needs setup, sending setup\r\n",cmd_data.remote_addr);
                    fnord_transmit_setup(cmd_data.remote_addr);

                    // warten bis eine Besttigung eintrifft wenn diese nicht kommt wird
                    // kein datenframe gesendet
                    f_tx_status = fnordnet_tx_status();
                    debug_printf("FNORDNET TX STATUS: %d\r\n",f_tx_status);
                }
                else
                {
                    f_tx_status = 1;
                }

                // Absender ist erreichbar
                if ( f_tx_status == 1)
                {
                    debug_printf("Target %d no setup needed\r\n",cmd_data.remote_addr);

                    // Also von remote angefordertes Kommando ausfuehren
                    (*cli_cmd[cmd_data.cmd_id].func)(&cmd_data);

                    // Resultat zurueckschicken an den absender
                    cmd_data.sub_command = cmd_data.cmd_id;
                    cmd_data.cmd_id = 255;

                    // Komplette cmd_data Stuktur (mit ergebnisdaten) an den absender schicken
                    // 11 Byte byte 1 = remote address wird nicht mit uebertragen
                    debug_printf("Transmitting result to %d\r\n",cmd_data.remote_addr);
                    fnord_transmit_frame(cmd_data.remote_addr,&cmd_data.raw[1],11);
                }
            }
        }
#endif

#if RFM69_OOK == 1
    //OOK Senden

    // Fifo Status
    //uint8_t status = rfm69_getreg(0x28);
    //debug_printf("Reg: 0x%02x",status);

    if (rfm_mode == 17)
    {
        uint8_t fifo[40];
        rfm69_read_fifo(fifo,40);
        for (i = 0 ; i < 40; i++)
        {
            //debug_printf("0x%02x ",fifo[i]);
            display_binary(fifo[i]);
        }
        debug_printf("\r\n");
        rfm_mode = 0;
    }

    if (ook_repeat_count > 0 && rfm_mode == 19)
    {
        ook_send_message();
        ook_repeat_count--;
        debug_printf("Sending OOK: %d \r\n",ook_repeat_count);
    }
    else if(ook_repeat_count == 0 && rfm_mode == 19)
    {
        rfm_mode = 0;
        // rfm69 in normalne FSK RX Modus versetzen
        rfm69_setmode(0); // Packet mode
        rfm69_setmodulation(0); // 0=FSK 1=OOK

        rfm69_setreg(0x2c,0x00); // Preamble (0xAA) count MSB = 0
        rfm69_setreg(0x2d,0x03); // Preamble (0xAA) count LSB = 3

        rfm69_setbaudrate(fnord_config.baudrate);

        //Sync Word setzen
        rfm69_setreg(0x2e,0x88); // UseSyncWord=ON, SyncWordLen=2Byte, FillFifo when SYNC Match
        rfm69_setreg(0x2f,0x2d); // Sync Word MSB
        rfm69_setreg(0x30,0xd4); // Sync Word LSB

        rfm69_setreg(0x37,0x80); //VarLenPacket DontCalcCRC // Seite 54

        // Normalen RX Mode
        rfm69_set_rxtx_mode(0);
        debug_printf("its code send\r\n");
    }
#endif

    } // ende while(1)
}

/* Diese Funktion dient dazu direkt im Jackalope ein Programm zur Steuerung
 * extern angeschlossener Einheiten unterzubringen, so dass es nicht ntig
 * ist die Steuerung ber die sehr langsame serielle schnittstelle laufen zu lassen
 * Alle Kommandos knnen weiterhin auf der seriellen schnittstelle ausgefhrt werden
 */
void exec_internal_programm(void)
{
    /* Example for an internal programm, switch outputs acording to an adc input
     * this programm is executed when mem 0 is set to 1 via cmdline  setmem 0 1
     * the mem value can be stored to eeprom with command savemem starting this
     * programm on powerup
     *
     * The exec_cmd routine is designed to use the cli text commands this makes it easy
     * to uitilze the commands
     */

    /* solar Control CCCamp2019*/

    /* 1. Schauen ob die Solarspannung > 10 Volt */
    /* 2. Wenn Ja DACOUT auf 0 und Relais 1 einschalten um den Wandler zu starten */

    /* 3. Regelung starten, Ausgangsstrom erhhen bis die Spannung auf < 10 Volt fllt dann 3 Schritte zurckregeln */

    /* 4. Wenn die Spannung bei 9.5 Volt ist, Ausgang abschalten */
    /* 5. Wenn die Spannung bei 9 Volt ist Wandler abschalten */

    /* Neuer Algo: 3.6.2022
     * 1. Ist die Leerlauf Spannung berhaupt hoch genug um einzuschalten
     * 2. Wenn ja Ausgangsleistung auf Minimum und einschalten
     * 3. Test Sturzt die Spannung am Solarmodul ab (4x messen und deltas zwichen den Messungen bilden wenn die immer grsser werden .. dann ja)
     * 4. Wenn ein Absturz vorliegt Ausgansleistung reduzieren wenn schon minimum einschaltschwelle(1) erhhen
     * 4.1 Ist die kritische minimalspannung erreicht -> wandlerleistung auf minimum
     * 4.2 Wieviel die Ausgangsleistung reduziert wird hngt von der strke des Absturzes ab und vom letzt angelegten leistungswert
     *
     * 5. Wenn kein Absturz vorliegt Leistung um STEP erhhen
     * 6. Weiter zu 3
     * STEP kann sich zusammensetzen aus Letzten stabilen Ausgangsleistungswert und Ausgangsleistungswert der einen Absturz verursachte
     *  z.B Differenz/2
     * Wenn der aktuelle Ausgangsleistungswert den letztmaligen Absturzwert berschreitet wird STEP erhht
     *
     *
     * Was passiert wenn mhrfach hintereinander die gleiche Schwellenausgangsleitung fr den Absturz gemessen wurde?
     * Es wre dann sinnvoll nur noch in kleinsten schritten zu regeln und den leerlaufspannungswert in
     * die Tabelle die Leerlaufspannungen zu Ausgangsleistungen zuordnet einzutragen
     */
#if SOLAR_REGULATOR_VERSION == 2
    int16_t cutoff_voltage; /* Relais AUS */
    int16_t start_voltage; /* Relais AN, system starten */
    int16_t max_dac_value; /* max Ausgangsstrom / leistung */
    int16_t min_dac_value; /* min dac value */
    int16_t voltageloss_level_threshold; /* swelle bei der angenommen wird dass die solar spannung zusammen bricht */
    /* anzahl der AD messungen die gemittelt werden*/
    int16_t adc_avgcount;

    /* Arbeitsvariablen */
    static uint8_t status = 0; // Status was macht das system grade / warten auf solarvoltage / regeln etc
    static int16_t act_solarvoltage = 0;
    static int16_t act_dac_value = 0;
    static int16_t last_ok_dac_value = 0;
    static int16_t last_ok_voltage_value = 0;
    static int16_t max_power_work_voltage = 0;
    static int16_t dac_start = 0;
    static uint16_t loops = 0;
    int16_t old_solarvoltage = 0;
    int16_t battery_voltage = 0;

    //loops++;

    start_voltage       = exec_cmd("getmem",5);
    cutoff_voltage      = exec_cmd("getmem",6);
    adc_avgcount        = exec_cmd("getmem",7);
    min_dac_value       = exec_cmd("getmem",8);
    max_dac_value       = exec_cmd("getmem",9);
    voltageloss_level_threshold = exec_cmd("getmem",10);

    last_ok_voltage_value = cutoff_voltage;

    /* Variable zur Triggerung der Ausgabe von Debug info */
    static int16_t delay = 0;

    act_solarvoltage = exec_cmd("adcavgin",1,adc_avgcount);

    /* Wenn die Spannung vom Solarpanel zu gering ist Relais aus, das wars */
    if (act_solarvoltage <= cutoff_voltage)
    {
        exec_cmd("dacout",1,0);
//        exec_cmd("out",1,0);
//        exec_cmd("out",2,0);
        status = 0;
        out_printf("  CUT OFF   \r\n");
//        loops = 0;
        dac_start = ( act_dac_value - 25);
        last_ok_dac_value = dac_start;
        max_power_work_voltage = 0;
    }

    /* Spannung vom Solarpanel bersteigt die untere Schwellspannung
     * EINSCHALTEN und Regelung starten
     */
    if (act_solarvoltage >= start_voltage && status == 0)
    {
        exec_cmd("dacout",1,0);
//        exec_cmd("out",1,1);
//        exec_cmd("out",2,1);
        if (dac_start < min_dac_value)
        { dac_start = min_dac_value; }
        act_dac_value = dac_start;
        exec_cmd("dacout",1,act_dac_value);
        max_power_work_voltage = 0;
        status = 1; // Regelung starten
        out_printf("  START  \r\n");
    }

    /* Regelung */
    /* Die Ausgangsleistung wird langsam erhht bis die spannung zusammenbricht dann zurck zum letzten ok wert*/
    if ( status == 1 || status == 2)
    {
        uint8_t i = 0;
        uint8_t max_i = 10;

        /* normaler betrieb leistung steigern */
        /* Spter nur dann die Leistung steigern wenn die Solarspannung ber der letzten ok spannung liegt */
        //if ( status == 1 && act_solarvoltage > max_power_work_voltage)
        if ( status == 1  && act_solarvoltage > start_voltage )
        {
            /* Ausgangsleistung steigern */
            act_dac_value++;

            if (act_dac_value > max_dac_value)
            { act_dac_value = max_dac_value; }

            exec_cmd("dacout",1,act_dac_value);

            out_printf(" UPDAC:%04d ",act_dac_value);

            max_i = 30;
        }
        else
        {
            out_printf(" SV:%03d  MAX-PW-SV:%03d ",act_solarvoltage,max_power_work_voltage);
        }

        /* ausgangsleistung war zu hoch alte ok leistung wurde eingestellt berwachen ob
         * die Ausgangsleistung jetzt ok ist fr einen lngern Zeitraum als normal
         */
        if ( status == 2)
        {
            max_i = 100;
        }

        int16_t probe_a = exec_cmd("adcavgin",1,adc_avgcount);

        /* schauen ob die Solarspannung stabil bleibt */
        for ( i = 1; i < max_i ; i++ )
        {
            int16_t probe_b = exec_cmd("adcavgin",1,adc_avgcount);

            //out_printf("TEST: IS:%03d REF:%03d\r\n",probe_b,probe_a);
            if (probe_b < (probe_a - 3) )
            {
                /* das war zu viel Ausgangsleistung
                 * Alten wert wieder einstellen und vermerken dass wir an der grenze waren
                 */
                exec_cmd("dacout",1,0);
                out_printf("  OVERLOAD DAC VALUE: %04d OK VALUE %04d\r\n",act_dac_value,last_ok_dac_value);
                last_ok_dac_value -= 2;
                if ( last_ok_dac_value < min_dac_value )
                { last_ok_dac_value = min_dac_value; }
                act_dac_value = last_ok_dac_value;
                exec_cmd("dacout",1,act_dac_value);
                max_power_work_voltage = 0;
                status = 2;
                break;
            }
        }

        /* wenn die Spannung stabil bleibt dann merken */
        if ( status == 1 )
        {
            last_ok_dac_value = act_dac_value;
            last_ok_voltage_value = exec_cmd("adcavgin",1,adc_avgcount);
            out_printf(" OK V:%03d DAC: %04d\r\n",last_ok_voltage_value,last_ok_dac_value);
        }

        /* wenn es eine berlast gab die stabilisiert wurde dann die Spannung merken
         * die zum stabilen Zustand gehrt. Die Regelung kann dann die Leistung erhhen wenn die
         * Solarspannung diese Spannung berschreitet
         */
        if (max_i == 100)
        {
            max_power_work_voltage = exec_cmd("adcavgin",1,adc_avgcount);
            status = 1;
            out_printf(" MAX POWER VOLTAGE \r\n",max_power_work_voltage);
        }
    }

    /* Regelung */
    if ( status == 1 && 1 == 0)
    {
        /* Testen ob die Solarzelle die Leistung liefern kann oder ob die Solarspannung
         * kontinuierlich zusammen bricht. Dazu mehrere Messugen vornehmen und diese miteinander verglichen
         * wird die differnz zwischen den Messungen immer grsser fllt die Spannung
         */
        int16_t probe_a = exec_cmd("adcavgin",1,adc_avgcount);
        int16_t probe_b = exec_cmd("adcavgin",1,adc_avgcount);
        int16_t probe_c = exec_cmd("adcavgin",1,adc_avgcount);
        int16_t probe_d = exec_cmd("adcavgin",1,adc_avgcount);

        // Berechnung der Differnezen im Zeitverlauf mit einer ber Gewichtung der spteren Werte
        int16_t voltageloss_level = (probe_a - probe_b) * 1 + (probe_b - probe_c) * 2 + (probe_c - probe_d) * 4;

        if (voltageloss_level >= voltageloss_level_threshold)
        {
            /* Solarspannung bricht zusammen ausgangsleistung muss abgesenkt werden */
            act_dac_value -= 20;
        }
        else
        {
            act_dac_value++;
        }

        if (act_dac_value > max_dac_value)
        { act_dac_value = max_dac_value; }

        exec_cmd("dacout",1,act_dac_value);
        if ( voltageloss_level != 0)
        { out_printf("  LOSS: %04d",voltageloss_level); }
    }
#endif

#if SOLAR_REGULATOR_VERSION == 1
    int16_t cutoff_voltage; /* Relais AUS */
    int16_t start_voltage; /* Relais AN, system starten */

    int16_t max_dac_value; /* max Ausgangsstrom / leistung */
    int16_t min_dac_value; /* min dac value */
    int16_t max_delay; /* wert den der delay zhler maximal haben darf*/
    int16_t multi;   /* wert mit dem der spannungseinbuch multipliziert wird um den dac auszuregeln */

    /* anzahl der AD messungen die gemittelt werden*/
    int16_t adc_avgcount;

    static int16_t max_solar_voltage = 0;
    static int16_t delay = 0;
    static int16_t dac_value = 0;
    static int8_t start_flag = 0;
    static int16_t act_solarvoltage = 0;
    int16_t old_solarvoltage = 0;
    int16_t battery_voltage = 0;

    start_voltage       = exec_cmd("getmem",5);
    cutoff_voltage      = exec_cmd("getmem",6);
    adc_avgcount        = exec_cmd("getmem",7);
    min_dac_value       = exec_cmd("getmem",8);
    max_dac_value       = exec_cmd("getmem",9);
    multi               = exec_cmd("getmem",10);

    max_delay = 50;

    delay++;
    out_printf("A");

    battery_voltage  = exec_cmd("adcavgin",2,adc_avgcount);
    exec_cmd("setmem",1,battery_voltage);
    old_solarvoltage = exec_cmd("adcavgin",1,adc_avgcount);
    act_solarvoltage = exec_cmd("adcavgin",1,adc_avgcount);
    out_printf("B");

    if (delay > max_delay)
    {
        delay = 0;
        if ( start_flag != 2 )
        {
            //out_printf("act: %03d old: %03d D: %03d DACOUT: %04d \r\n",act_solarvoltage,old_solarvoltage,(old_solarvoltage-act_solarvoltage),dac_value);
            out_printf("act: %03d bat: %03d\r\n",act_solarvoltage,battery_voltage);
        }
    }


    if (act_solarvoltage > max_solar_voltage)
    {
        max_solar_voltage = act_solarvoltage;
    }

    /* Relais an, es ist genug spannung da */
    if (act_solarvoltage >= start_voltage && start_flag == 0 && delay == max_delay )
    {
        // Ausgang auf mimimalen Ausgangsstrom
        dac_value = min_dac_value;
        exec_cmd("dacout",1,dac_value);
        // Eingangsrelais AN
        //exec_cmd("out",1,1);
        exec_cmd("out",2,1);
        start_flag = 1;
        delay = 0;
        out_printf("C");
    }

    if ( start_flag == 1 && delay == max_delay)
    {
        delay = 0;
        start_flag = 2;
        out_printf("D");
    }

    /* vom eingang kommt zu wenig, alles abschalten */
    /* oder trotz dass der ausgangsstrom maximal ist ist die solarspannung maximal -> dc/dc wandler resetten */
    /* if ( (solarvoltage <= cutoff_voltage && start_flag == 1) || (dac_value >= max_dac_value && solarvoltage >= (max_solar_voltage-20) && start_flag == 1) ) */
    if ( act_solarvoltage <= cutoff_voltage && start_flag > 0  )
    {
        // Ausgang auf mimimalen Ausgangsstrom
        dac_value = min_dac_value;
        exec_cmd("dacout",1,dac_value);
        // Eingangsrelais AUS
        //exec_cmd("out",1,0);
        exec_cmd("out",2,0);
        start_flag = 0;
        delay = 0;
        out_printf("E");
    }

    /* Regelung */
    if ( start_flag == 2 )
    {
        /* Nur wenn der AD Wandler maximal um 1 Zhler fllt darf nach oben geregelt werden */
        if ( act_solarvoltage >= old_solarvoltage )
        {
            if (dac_value < max_dac_value && act_solarvoltage > cutoff_voltage)
            {
                dac_value++;
                exec_cmd("dacout",1,dac_value);
                out_printf("*");
            }
        }
        else
        {
            int16_t reduction = ( old_solarvoltage - act_solarvoltage) ;
            out_printf("%02d",reduction);
            reduction = reduction * multi;
            dac_value = dac_value - reduction;

            if (dac_value < min_dac_value)
            { dac_value = min_dac_value; }

            exec_cmd("dacout",1,dac_value);
            // out_printf("! %04d %03d ",reduction,dac_value);
        }
    }
#endif

    /* ausgang abschalten um ein totalzusammenbruch wegen zu geringer eingangsleistung zu verhindern */
    /*
    if (solarvoltage <= output_off_voltage)
    {
        dac_value = min_dac_value;
        exec_cmd("dacout",1,dac_value);
    }
    */

    /* system ist ausgeregelt */
    /*
    if (solarvoltage <= min_ok_voltage && dac_value > min_dac_value)
    {
        dac_value = dac_value - reduction;
        exec_cmd("dacout",1,dac_value);
    }
    */

    /* ausgangsleistung erhhen, und sehen was passiert */
    /*
    if (start_flag == 1 && delay > max_delay)
    {
        delay = 0;
        if (dac_value < max_dac_value && solarvoltage > min_ok_voltage)
        {
            if (solarvoltage > 650)
            { dac_value += 2; }
            else
            {   dac_value++; }
            exec_cmd("dacout",1,dac_value);
        }
        out_printf("Vin: %03d DACOUT: %04d MAX: %03d\r\n",solarvoltage,dac_value,max_solar_voltage);
    }
    */
/*
    //int16_t exec_cmd(uint8_t cmd_id, int16_t value1, int16_t value2)
    int16_t a;

    a = exec_cmd("adcin",1);
    //if (a < merker.wert[1])
    if (a < 300)
    {
        exec_cmd("out",2,1);
    }
    else
    {
        exec_cmd("out",2,0);
    }

    //if (a < merker.wert[1])
    if (a < 400)
    {
        exec_cmd("out",1,1);
    }
    else
    {
        exec_cmd("out",1,0);
    }

*/
}

/* Alles timer port etc... initialisieren */
void init(void)
{
    /* Power reduction Register */
    PRR = 0; // enable all units

    /* SMCR Sleep Mode Control Register Seite 37 */
    SMCR = 0x01; // asm sleep Instruction schaltet den IDLE Mode ein

    // Timer0 fuer UI display

    // FastPWM Mode
    TCCR0A = _BV(WGM01) | _BV(WGM00);

    // prescaler = 1024 FastPWM mode mit OCR0A als TOP
    TCCR0B |=  _BV(WGM02) | _BV(CS02) | _BV(CS00);

    // Output compare A match Interrupt an
    TIMSK0 |= _BV(OCIE0A);

    OCR0A = 195; // ca. 10ms pro Interrupt
    TCNT0 = 0;

    // IO1 bis IO6
    //DDRD = 0b00001000;
    //PORTD = 0b00001000;
}

void adcinit(void)
{
    /* Prescaler 64 */
    ADCSRA &= ~_BV(ADPS0);
    ADCSRA |= _BV(ADPS1);
    ADCSRA |= _BV(ADPS2);

    /* Refernz Spannung = intern VCC */
    ADMUX |= _BV(REFS0);

    if ( fnord_config.adcref == 1)
    {
        /* Referenz 1.1 Volt internal reference */
        ADMUX |= _BV(REFS1);
    }

    /* Single Conversion Mode */
    ADCSRA &= ~_BV(ADATE); // Auto Trigger Disable

    /* Start Conversion */
    ADCSRA |= _BV(ADEN); // ADC Enable
}

void set_cfg_defaults(void)
{
    fnord_config.local_address = 10;
    fnord_config.channel = 12;
    fnord_config.adcref = 5;
    fnord_config.json = 0;

#if RFM12_ENABLE == 1
    fnord_config.rssi = 2;
    fnord_config.baudrate = 9578;
    fnord_config.txpower = 7;
#endif

#if RFM69_ENABLE == 1
    fnord_config.rssi = 92; // -92 dB
    fnord_config.baudrate = 12800; // even divisor
    fnord_config.txpower = 8; //8 =-10dBm = 100 uWatt, min = 0 = -18dBm = 15uWatt
    fnord_config.rfmzin = 50;
#endif
}

// Wird alle 10ms angesprungen
ISR(TIMER0_COMPA_vect)
{
    if (systemcounter == 10)
    {
        ui_display_flag = 1;
        systemcounter=0;
    }
    else
    {
        systemcounter++;
    }
#if FNORD_NET_ENABLE == 1
    fnord_countup_silence_count();
#endif
}

#if RFM69_OOK == 1
ISR(TIMER1_OVF_vect)
{
    //serial_output("X");
    ti1_ovf_cnt++;
}

// CodeCapture 433MHZ OOK Funkfernbedienungen ueber PB0
#if RFM69_OOK_CAPTURE == 1
ISR(PCINT0_vect)
{
    uint8_t rssi = 0;
    uint8_t pinlevel;
    uint16_t timerval;

    pinlevel = (PINB & 0b00000001);
    if ( pinlevel == pinlevel_pb0 )
    { return; }

    pinlevel_pb0 = pinlevel;

    // RSSI Checken DIO0 einlesen
    rssi = (PINC & 0b00100000);

    // Wenn RSSI nicht ausreichend Operation des einlesens abbrechen
    if (rssi == 0) { return; }

    timerval = TCNT1;
    TCNT1 = 0;

    if (timerval < 10)
    {
        return;
    }

    if (timerval > 50000)
    {
        return;
    }

    ooktime_count++;

    //timerval = (timerval << 5); // *32
    //timerval = timerval / 10;

    // Steigende oder fallende Flanke PB0
    pinlevel = (PINB & 0b00000001);

    /*
        rfm69_setreg(0x23,0x01);//Start RSSI Detect
        while ( (rfm69_getreg(0x23) & 0x02) == 0 ) { asm("nop"); }
        rssi = rfm69_getreg(0x24);
        debug_printf("%u,%u ",timerval,pinlevel);
        return;
    */

    // Steigende Flanke
/*
    if (pinlevel == 1)
    {
        // RSSI Checken
        rssi = rfm69_getreg(0x27);
        //rssi &= 0x08;
        // RSSI Reset = Bit 3 = rssi auf 0 setzen damit es geloescht wird
        rfm69_setregmask(0x27,0x08,0x00);

        debug_printf("%u ",rssi);

        // RSSI Checken DIO0 einlesen
        //rssi = (PINC & 0b00100000);

        // Wenn RSSI nicht ausreichend Operation des einlesens abbrechen
        if (rssi == 0)
        {
            //serial_output("0");
            if (ook_capture == 1)
            {
                serial_output("X\r\n");
            }
            ook_capture = 0;
        }
        else
        {
            ook_capture = 1;
            //serial_output("1");
        }
    }
    else
    {
            //serial_output("2");
    }
*/
        // Timerstand ausgeben und timer resetten
        pinlevel = pinlevel ^ 0x01;
        //timerval = timerval * 3;
        debug_printf("%u,%u ",timerval,pinlevel);

    if (ooktime_count >= 600)
    {
        rfm69_diable_capture();
    }
}
#endif

// 433 MHz senderoutine
ISR(TIMER1_COMPA_vect)
{
        //cli();
        //sei();
}
#endif


/* USART0 Receive register contains valid data to be processed */
ISR(USART_RX_vect)
{
    char c;
    c = UDR0;

    if (c == 0x0a || c == 0x0d)     /* Zeilenende -> Kommando verarbeiten */
    {
        ui_cmd_ready = 1;
        command[ui_uart_rx_pointer] = 0x00;
        #if ANSI_DISPLAY_ENABLE == 0
        sprintf_P(uartbuf,PSTR("\r\n"));
        serial_output(uartbuf);
        #endif
        return;
    }

    if ( c < 0x20 ) /* steuerzeichen nicht in den puffer schreiben */
    { return ; }

    command[ui_uart_rx_pointer] = c;
    command[ui_uart_rx_pointer+1] = 0x00;

    if ( (c == 0x08 || c == 0x7f) )     /* backspace */
    {
        command[ui_uart_rx_pointer] = 0x00;
        if ( ui_uart_rx_pointer > 0)
        {
            ui_uart_rx_pointer--;
            command[ui_uart_rx_pointer] = 0x00;
            sprintf_P(uartbuf,PSTR("%c"),0x7f);
            serial_output(uartbuf);
        }
    }
    else
    {
        if ( ui_uart_rx_pointer < 30)
        {
            ui_uart_rx_pointer++;
        }
        #if ANSI_DISPLAY_ENABLE == 0
        sprintf_P(uartbuf,PSTR("%c"),c);
        serial_output(uartbuf);
        #endif
    }

/*
 *  Noch nicht so ganz fertig
    if (ui_uart_rx_pointer >= 2 && command[ui_uart_rx_pointer] == 65 && command[ui_uart_rx_pointer-1] == 91 && command[ui_uart_rx_pointer-2] == 27)
    {
        // Cursor Hoch
        //strcpy(command,command_old);
        //ui_uart_rx_pointer= ui_uart_rx_pointer_old;
        ui_uart_rx_pointer -= 3;
        command[ui_uart_rx_pointer] = 0x00;
    }
*/
}

/*
 * Diese Funktion prueft ob gerade was empfangen wird und verhindert damit das senden eines frames
 * gibt eine 0 zurck wenn gesendet werden kann und eine 1 wenn eine kollision vorliegen wrde
 * wie diese funktion das macht bleibt den treibern de funkmodule berlassen
 */
uint8_t fnord_check_collision(void)
{
#if RFM12_ENABLE == 1
    return 0;
#endif

#if RFM69_ENABLE == 1

    /*
    uint8_t reg = rfm69_getreg(0x27);
    debug_printf("TX COL 0x%02x\r\n",reg);
    return 0;
    */

    /*
    uint8_t act_rssi = rfm69_getrssi();
    debug_printf("TX COL: %u\r\n",act_rssi);
    if ( act_rssi > fnord_config.rssi)
    { return 1; debug_printf("TX COLLISSION\r\n");}
    else
    { return 0 ;}
    */


    static uint8_t fnord_collision_count = 0;

    if (fnord_collision_count >= FNORD_MAX_COLLISION_COUNT)
    {
        debug_printf("ERROR TOO MANY COLLISIONS, resetting module\r\n");

        // Standby modus damit die RX Flags in Reg 0x27 geloescht werden
        rfm69_set_rxtx_mode(2);

        asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop");

        // RX Mode wieder starten
        rfm69_set_rxtx_mode(0);

        asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop");
        fnord_collision_count=0;

        // eventuell hier mit return 0 beenden falls reg0x27 & 0x01 immernoch auf eine kollision hindeuten
    }

    uint8_t reg = rfm69_getreg(0x27);
    if ( (reg & 0x01) == 1)
    {
        //debug_printf("TX COLISSION 0x%02x\r\n",reg);
        debug_printf("TX COLISSION 0x%02x CNT: %d\r\n",reg,fnord_collision_count);
        fnord_collision_count++;
        return 1;
    }
    else
    {
        fnord_collision_count = 0;
        return 0;
    }
#endif
}

#if RFM12_ENABLE == 1 || RFM69_ENABLE == 1
void fnord_tx_request(uint8_t *fnord_data,uint8_t bytelen)
{
#if RFM12_ENABLE == 1
    rfm12_tx_bytestream(fnord_data,bytelen);
#endif
#if RFM69_ENABLE == 1
    rfm69_tx_bytestream(fnord_data,bytelen);
#endif
}
#else
    #if FNORD_NET_ENABLE == 1
    void fnord_tx_request(uint8_t *fnord_data,uint8_t bytelen)
    {
    // Nothing to do RFM12/RFM69 not enabled
    debug_printf("No RF Module, no transmit!\r\n");
    }
    #endif
#endif

#if FNORD_NET_ENABLE == 1
uint8_t fnordnet_tx_status()
{
    uint32_t i = 0;
    #define COUNT 300000

    do
    {
        fnord_process();
        i++;
    } while (fnord_get_tx_status() == 0 && i < COUNT);

    if ( i >= COUNT)
    {
        return 20;
    }
    else
    {
        return fnord_get_tx_status();
    }
}
#endif

void uart_init(void)
{
    /* 115200 BPS at 20MHz systemclock*/
    #define UBRR 10

    /* 38400 BPS at 8MHz systemclock*/
//    #define UBRR 12
    /* Baudrate setzen */
    UBRR0H = HI8(UBRR) ;
    UBRR0L = LO8(UBRR) ;
    /* Enable receiver and transmitter */
    UCSR0B = _BV(RXEN0) | _BV(TXEN0) ;

    /* Set data format 8 DataBits 2 StopBits */
    //UCSR0C = _BV(USBS0) | _BV(UCSZ00) | _BV(UCSZ01);

    /* Set data format 8 DataBits 1 StopBits */
    UCSR0C = _BV(UCSZ00) | _BV(UCSZ01);

    /* enable receive interrupt */
    UCSR0B |= _BV(RXCIE0) ;
}

void serial_output(char *msg)
{
  char *i ;
  uint8_t counter = 0;
  i = msg ;

  while(*i > 0 && counter < 90)
  {
    uart_sendbyte(*i) ;
    i++;
    counter++;
  }

  if (counter == 90)
  {
    sprintf_P(uartbuf,PSTR("UART BUFFFER OVERFLOW\r\n"));
    i = uartbuf;
    while(*i > 0)
    {
        uart_sendbyte(*i) ;
        i++;
    }
  }
}

void display_binary(uint8_t data)
{
    uint8_t bv = 0x80;
    uint8_t i;
    for(i = 0; i< 8; i++)
    {
        if( data & bv)
        { debug_printf("1");}
        else
        { debug_printf("0");}
        bv = (bv >> 1);
    }
}

void uart_sendbyte(char data)
{
    /* wait till tx buffer is empty */
    while ( ! (UCSR0A  & _BV(UDRE0)) )
    {}
    /* send byte */
    UDR0 = data;
}

void spi_init(void)
{
    // Check for Bootloader enable Jumper set
    /*
    uint8_t pinb;
    DDRB=0; //All input
    PORTB = 0x02; // PB1 Pullup on
    for (pinb = 0 ; pinb < 200 ; pinb++)
    { asm("nop"); } // little delay
    pinb = PINB;
    PORTB = 0x00; // PB1 Pullup off

    // RFM12_CS pulled to low by jumper
    if ( (pinb & 0x02) == 0)
    {
        // Jmp to bootloader
        out_printf("Detected bootloader jumper set\r\n");
        out_printf("Starting bootloader...\r\n");
        jump_to_bootloader();
    }
    */

    DDRC |= _BV(PC4); // DAC_CS
    PORTC |= _BV(PC4); // INAKTIV

    DDRB |=  _BV(PB5) |  _BV(PB3) | _BV(PB2) ; // MOSI und SCK und SS = OUTPUT

//  SPCR = _BV(SPE) | _BV(MSTR) | _BV(SPR1) | _BV(SPR0); // Master Mode Clock Diver 128
    SPCR = _BV(SPE) | _BV(MSTR) | _BV(SPR0); // Master Mode Clock Diver 16
//  SPCR = _BV(SPIE) | _BV(SPE) | _BV(MSTR);

    // Deselect SPI devices
    // RFM12 Chip Select
    DDRB |= _BV(PB1);
    PORTB |= _BV(PB1);

    // DAC deselct
    DDRC |= _BV(PC4);
    PORTC |= _BV(PC4);
}

uint8_t spi_transfer_byte(uint8_t spi_tx)
{
    SPDR = spi_tx;
    asm("nop"); // DELAY
    /* Wait for Transmission to complete */
    uint16_t i = 0;
    while(  (SPSR & _BV(SPIF)) == 0 ) { i++; };
    //debug_printf("SPI Wait: %u\r\n",i);
    return SPDR;
}


/* --------------------- COMMAND LINE INTERFACE ----------------------- */

void ui_display(void)
{
#if ANSI_DISPLAY_ENABLE == 1
    uint8_t pos = 0 ;
    ansi_locate(1,1);
    debug_printf("CMD: %s          ",command);
    ansi_locate(2,1);
    //debug_printf("Message: %s          ",message);

#if FNORD_NET_ENABLE == 1
    ansi_locate(3,1);
    debug_printf("Fnord status TX: %d RX: %d   ",fnord_get_tx_status(),fnord_get_rx_status());
#endif
#endif
}

/* ANSI Funktionen */
/* Diese Funktion loescht den Bildschirm */
void ansi_cls(void)
{
    sprintf_P(uartbuf,PSTR("%c[2J"),27);
    serial_output(uartbuf);
    ansi_locate(1,1);
}

/* Diese Funktion setzen den cursor an die angegebene Position */
void ansi_locate(int x, int y)
{
    sprintf_P(uartbuf,PSTR("%c[%d;%dH"),27,x,y);
    serial_output(uartbuf);
}

/* Diese Funktion setzt vordergrund und Hintergrund Farbe
 * 0=Black 1=red 2=green 3=yellow 4=blue 5=mangenta 6=cyan 7=white
 */
void ansi_color(int fg, int bg)
{
    fg = fg + 30 ;
    bg = bg + 40 ;
    sprintf_P(uartbuf,PSTR("%c[%d;%dm"),27,fg,bg);
    serial_output(uartbuf);
}

//int16_t exec_cmd(const char *cmd_name, int16_t value1, int16_t value2)
//{
//    uint8_t cmd_id = 0 ;
//    cmd_data.value[0] = value1;
//    cmd_data.value[1] = value2;
//
//    while (cli_cmd[cmd_id].parametercount > 0)
//    {
//        /* wenn der Text des Kommandos uebereinstimmt dann ausfuehren */
//        if ( strncmp_P(cmd_name,cli_cmd[cmd_id].cmd_text,strlen_P(cli_cmd[cmd_id].cmd_text)) == 0)
//        {
//            (*cli_cmd[cmd_id].func)(&cmd_data);
//            /* es ist ein Fehler aufgetreten */
//            if ( cmd_data.err > 0 )
//            {
//            }
//        }
//    cmd_id++;
//    }
//
//    return cmd_data.value[1];
//}

int16_t exec_cmd(const char *cmd_name, int16_t value, ... )
{
    uint8_t cmd_id = 0 ;
    va_list args;
    va_start(args, value);

    cmd_data.value[0] = value;
    cmd_data.value[1] = (int16_t)va_arg(args, int16_t);
    cmd_data.value[2] = (int16_t)va_arg(args, int16_t);
    cmd_data.value[3] = (int16_t)va_arg(args, int16_t);
    va_end(args);

    while (cli_cmd[cmd_id].parametercount > 0)
    {
        /* wenn der Text des Kommandos uebereinstimmt dann ausfuehren */
        if ( strncmp_P(cmd_name,cli_cmd[cmd_id].cmd_text,strlen_P(cli_cmd[cmd_id].cmd_text)) == 0)
        {
            (*cli_cmd[cmd_id].func)(&cmd_data);
            /* es ist ein Fehler aufgetreten */
            if ( cmd_data.err > 0 )
            {
            }
        }
    cmd_id++;
    }
    return cmd_data.value[1];
}


void process_cli_command(void)
{
    ui_cmd_ready = 0;
    ui_uart_rx_pointer = 0;
    char cmd[40];
    cmd[0] = 0x00;
    uint8_t argcnt;
    uint8_t cmd_id;

    //debug_printf("Command: %s\r\n",command);

    //http://www.nongnu.org/avr-libc/user-manual/group__avr__stdio.html#ga67bae1ad3af79809fd770be392f90e21

    cmd_data.value[0] = 0;
    cmd_data.value[1] = 0;
    cmd_data.value[2] = 0;
    cmd_data.value[3] = 0;

    //argcnt = sscanf_P(command,PSTR("%s %d %d %d"),cmd,&value1,&value2,&value3);
    argcnt = sscanf_P(command,PSTR("%s %i %i %i %i"),cmd,&cmd_data.value[0],&cmd_data.value[1],&cmd_data.value[2],&cmd_data.value[3]);
    cmd_data.err = 0;

    //debug_printf("Parameter V1: %d V2: %d V3: %d\r\n",cmd_data.value[0],cmd_data.value[1],cmd_data.value[2]);

    cmd_id=0;
    while (cli_cmd[cmd_id].parametercount > 0)
    {
        //debug_printf("CMD ID %d\r\n",cmd_id);
        /* wenn der Text des Kommandos und die Anzahl der Parameter uebereinstimmen
         * dann ausfuehren */
        //   && strlen(command) == strlen_P(cli_cmd[cmd_id].cmd_text)
        if ( strncmp_P(cmd,cli_cmd[cmd_id].cmd_text,strlen_P(cli_cmd[cmd_id].cmd_text)) == 0)
        {
            if (argcnt == cli_cmd[cmd_id].parametercount)
            {
                if (fnord_remote_address == 0 || cli_cmd[cmd_id].always_local == 1)
                {
                    // Kommando lokal ausfuehren
                    (*cli_cmd[cmd_id].func)(&cmd_data);

                    if (fnord_config.json == 0)
                    {
                        if ( cmd_data.err < 2 )
                        {
                            sprintf_P(uartbuf,cli_cmd[cmd_id].return_text,cmd_data.value[0],cmd_data.value[1],cmd_data.value[2],cmd_data.value[3]);
                            serial_output(uartbuf);
                            sprintf_P(uartbuf,PSTR("\r\n"));
                            serial_output(uartbuf);
                        }
                        else
                        {
                            sprintf_P(uartbuf,errormessage[cmd_data.err],cmd_data.value[0],cmd_data.value[1],cmd_data.value[2],cmd_data.value[3]);
                            serial_output(uartbuf);
                            sprintf_P(uartbuf,PSTR("\r\n"));
                            serial_output(uartbuf);
                        }
                    }
                    else if (fnord_config.json == 1)
                    {
                        if ( cmd_data.err < 2)
                        {
                            sprintf_P(uartbuf,PSTR("{ \"value1\":\"%d\",\"value2\":\"%d\",\"value3\":\"%d\",\"value4\":\"%d\" }"),cmd_data.value[0],cmd_data.value[1],cmd_data.value[2],cmd_data.value[3]);
                            serial_output(uartbuf);
                            serial_output((char*)"\r\n");
                        }
                        else
                        {
                            if ( cmd_data.err > 14 ) {cmd_data.err = 14;}
                            sprintf_P(uartbuf,PSTR("{ \"error\":\"%u\", \"errortext\":\""),cmd_data.err);
                            serial_output(uartbuf);
                            sprintf_P(uartbuf,errormessage[cmd_data.err],cmd_data.value[0],cmd_data.value[1],cmd_data.value[2],cmd_data.value[3]);
                            serial_output(uartbuf);
                            serial_output((char*)"\" }\r\n");
                        }
                    }
                    else
                    {
                        out_printf("JSON / TXT output cfg not set\r\n");
                    }
                    break;
                }
#if FNORD_NET_ENABLE == 1
                else
                {
                    // Kommando remote ausfuehren
                    cmd_data.remote_addr = fnord_remote_address;
                    cmd_data.cmd_id = cmd_id;
                    cmd_data.sub_command = 0;
                    cmd_data.err = 0x55; // 0b01010101 normalerweise =0 spezieller wert fuer gute funkuebertragung wird auf empfaengerseite verworfen
                    // 11 Byte cmd_data.remote_addr wird nicht uebertragen
                    fnord_transmit_frame(cmd_data.remote_addr,&cmd_data.raw[1],11);
                    break;
                }
#endif
            }
            else
            {
                // Kommandotext gefuden aber die anzahl der parameter stimmt nicht
                // Hilfetext ausgeben
#if HELP_ENABLE == 1
                if (fnord_config.json == 0)
                {
                    sprintf_P(uartbuf,PSTR("ERROR Invalid parameters usage:  "));
                    serial_output(uartbuf);

                    sprintf_P(uartbuf,cli_cmd[cmd_id].help_text);
                    serial_output(uartbuf);

                    sprintf_P(uartbuf,PSTR("\r\n"));
                    serial_output(uartbuf);
                    break;
                }
#endif
            }
        }
        cmd_id++;
    }

    // Alle kommandos durchlaufen aber kein passendes gefunden
    if ( cli_cmd[cmd_id].parametercount == 0 && strlen(cmd) > 2)
    {
        if (fnord_config.json == 0)
        {
            out_printf("ERROR: invalid command type help for help\r\n");
        }
        else if (fnord_config.json == 1)
        {
            out_printf("{ \"error\":\"1\", \"errortext\": \"invalid command\" }\r\n");
        }
        else
        {
            out_printf("JSON / TXT output cfg not set\r\n");
        }
    }
}

/* Command functions */
void cmd_version(struct cmd_data_t *cmd)
{
    cmd->value[0] = MAJOR_VERSION;
    cmd->value[1] = MINOR_VERSION;
    cmd->err = 1;
}

void cmd_test(struct cmd_data_t *cmd)
{
    /* execute command */
    //debug_printf("executing test from: %d Data: %d\r\n",cmd->remote_addr,cmd->value[0]);

    /* send reply */
    cmd->value[0] = 111;
    cmd->value[1] = 222;
    cmd->value[2] = 333;
    cmd->value[3] = 444;
    cmd->err = 1;
}

void cmd_crc32(struct cmd_data_t *cmd)
{
    uint32_t crc;
    crc = 0;
//    const char data[] = { "123456789" };
//    crc = crc32_checksum(data, strlen(data) );
    out_printf("CRC: %08lx",crc);
}

void cmd_bootloader(struct cmd_data_t *cmd)
{
    out_printf("\r\n");
    jump_to_bootloader();
}


void cmd_memfree(struct cmd_data_t *cmd)
{
    int v;
    cmd->value[0] = (int16_t) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval);
}

#if DHT11_ENABLE == 1
void cmd_temp(struct cmd_data_t *cmd)
{
    cmd->err = 1;
    dht11_read(&cmd->value[0],&cmd->value[1],&cmd->err);
}
#endif

#if RFM69_DEBUG_CMDS == 1 && RFM69_ENABLE == 1
void cmd_rfm(struct cmd_data_t *cmd)
{
    cmd->value[0] = rfm69_sendcmd(cmd->value[0],0,0);
}

void cmd_rfm_write(struct cmd_data_t *cmd)
{
    cmd->value[1] = rfm69_sendcmd(cmd->value[0], cmd->value[1],1);
}

void cmd_fifow(struct cmd_data_t *cmd)
{
    uint8_t daten[20] = "12345678901234567890";
    rfm69_write_fifo(daten, 20);
}

void cmd_fifor(struct cmd_data_t *cmd)
{
    uint8_t daten[20];
    rfm69_read_fifo(daten,20);
}
#endif

// Interntechno default Adresse setzen
//
//
void cmd_itsset(struct cmd_data_t *cmd)
{
    uint32_t bits;
    uint32_t tmp;

    if ( cmd->value[0] == 0 && cmd->value[1] == 0 )
    {
        cmd->value[1] = (int16_t)fnord_config.its_addr & 0x0000ffff;
        cmd->value[0] = (fnord_config.its_addr >> 16);
    }
    else
    {
        // First Value = first 16 Bits of address
        tmp = (uint32_t)cmd->value[0];
        bits = (tmp << 16);

        // Second value 10 Bits MSB for next 10 Bits of address
        tmp = (uint32_t)cmd->value[1];
        tmp &= 0xffc0; // nur die oberen 10 bit
        bits |= tmp;

        fnord_config.its_addr = bits;
    }
}

// TOOM Funksteckdosen
// http://www.forum-raspberrypi.de/Thread-selbstlernende-toom-funksteckdose-1919384-schalten
//
// http://www.forum-raspberrypi.de/Thread-selbstlernende-toom-funksteckdose-1919384-schalten
// https://www.sweetpi.de/blog/329/ein-ueberblick-ueber-433mhz-funksteckdosen-und-deren-protokolle
//
// http://www.mikrocontroller.net/topic/124084
//
// KORRETKTE REFERENZ 2022
// https://forum-raspberrypi.de/forum/thread/25230-selbstlernende-toom-funksteckdose-1919384-schalten/
//
// ----------------------------------------------------------------------------
// Intertechno Protokoll
// http://www.fhemwiki.de/wiki/Intertechno_Code_Berechnung#Selbstlernende_Intertechno_Funksteckdosen_.28z.B._ITR-1500.29
//
// NICHT TOOM
// 0-Bit    = 1T High + 3T Low + 1T High + 3T Low
// 1-Bit    = 1T High + 3T Low + 3T High + 1T Low
// Sync-Bit = 1T High + 31T Low
//
// 1T etwa 300 Mikrosekunden
// Baudrate setting = 3333 BPS = 0x25(Hi) + 0x80(Lo)
//

// TOOM Baumarkt Steckdosen schalten
#if RFM69_OOK == 1
void cmd_toom(struct cmd_data_t *cmd)
{
    uint32_t bits;
    uint32_t tmp;

    bits = 0;

    if (cmd->value[0] == 1) // TOOM Baumarkt
    { bits = 0xfc431c00 ; }
    else if (cmd->value[0] == 0)
    { bits = 0xf3e58c00; }
    else if (cmd->value[0] == 2)
    { bits = 0xca3d4000; }
    else if (cmd->value[0] == 3)
    { bits = 0xccf79000; }

    // Senden vorbereiten
    cli();

    //RFM69 in OOK Modus schalten
    rfm69_setreg(0x25,0x80); // disable DIO0
    rfm69_setmode(3); // Continous without sync
    rfm69_setmodulation(1); // 0=FSK 1=OOK

    // PB0 auf Ausgang level auf 0
    DDRB |=  _BV(PB0);
    PORTB &= ~_BV(PB0);

    // TX Mode
    rfm69_set_rxtx_mode(1);

    sei();

    // ---- Senden Anfang ----

    uint32_t tx;
    uint8_t i,j;
    uint8_t bit;

    debug_printf("OOK Send: ");

    // TOOM Baumarkt Funksteckdose 1919384


    for (j=0; j<4; j++)
    {
        ook_sendbit(2, 1); // sync senden
        tx = bits;
        for (i=0; i<24; i++)
        {
            tmp = tx & 0x80000000;
            if (tmp) {bit = 1;} else {bit = 0;}
            tx = (tx << 1);
            ook_sendbit(bit, 1);
            //debug_printf("%u,",bit);
        }
            //debug_printf("\r\n");
    }


/*
    for (j=0; j<4; j++)
    {
        ook_sendbit(2, 2); // sync senden
        tx = bits;
        for (i=0; i<24; i++)
        {
            tmp = tx & 0x80000000;
            if (tmp) {bit = 1;} else {bit = 0;}
            tx = (tx << 1);
            ook_sendbit(bit, 2);
            //debug_printf("%u,",bit);
        }
            debug_printf("\r\n");
    }
*/
    debug_printf("\r\n");

    // PB0 auf Eingang Pullup aus
    DDRB &=  ~_BV(PB0);
    PORTB &= ~_BV(PB0);

    //RFM69 in normalen modus schalten
    rfm69_setmode(0); // Packet mode
    rfm69_setmodulation(0); // 0=FSK 1=OOK
    rfm69_set_rxtx_mode(0); // RX Mode

    cmd->err = 0;

    //Senden Ende


} // ende cmd_toom


void cmd_ooksend(struct cmd_data_t *cmd)
{
    uint32_t bits;
    uint32_t tmp;

    // First Value = first 16 Bits of address
    tmp = (uint32_t)cmd->value[0];
    bits = (tmp << 16);

    // Second value 10 Bits MSB for next 10 Bits of address
    tmp = (uint32_t)cmd->value[1];
    tmp &= 0xffc0; // nur die oberen 10 bit
    bits |= tmp;

    // Wenn 0x0 0x0 als adresse angegeben wird die in fnord_config hinterlegte adresse nehmen
    if ( bits == 0 )
    { bits = fnord_config.its_addr; }

    // Third value high nibble (4Bits) = keypair 0x00 to 0xF0
    tmp = (uint32_t)cmd->value[2];
    tmp = (tmp >> 4);
    bits |= tmp;

    // Third Value low nibble = ON / OFF 00 = off, 01 = on 10= all off, 11= all ON -> 0x00 to 0x03
    tmp = (uint32_t)(cmd->value[2] & 0x0003);
    tmp = (tmp << 4);
    bits |= tmp;

    debug_printf("its bits: 0x%08lx\r\n",bits);


/*
    if (cmd->value[0] == 1) // TOOM Baumarkt
    { bits = 0xfc431c00 ; }
    else if (cmd->value[0] == 0)
    { bits = 0xf3e58c00; }
    else if (cmd->value[0] == 2)
    { bits = 0xca3d4000; }
    else if (cmd->value[0] == 3)
    { bits = 0xccf79000; }

    // Intertechno  // 26 Bit hauscode + 2Bit ein/aus + 4 bit tastenpaar
    else if (cmd->value[0] == 4)
    { bits = 0x00445611; }
    else if (cmd->value[0] == 5)
    { bits = 0x00445601; }
    else
    { bits = 0x00000000; }
*/

#if RFM69_OOKMODE == 1
    cli();
    //RFM69 in OOK Modus schalten
    rfm69_setreg(0x25,0x80); // disable DIO0
    rfm69_setmode(3); // Continous without sync
    rfm69_setmodulation(1); // 0=FSK 1=OOK

    // PB0 auf Ausgang level auf 0
    DDRB |=  _BV(PB0);
    PORTB &= ~_BV(PB0);

    // TX Mode
    rfm69_set_rxtx_mode(1);
    sei();
#endif

// https://github.com/Phunkafizer/RaspyRFM/blob/master/intertechno.py
// OOK senden ueber FIFO
#if RFM69_OOKMODE == 2
    ook_message = bits;
    rfm69_setbaudrate(3570); // Intertechno
    //rfm69_setbaudrate(2666); // TOOM Baumarkt

    rfm69_setmodulation(1); // 0=FSK 1=OOK

    //Preamble ausschalten
    rfm69_setreg(0x2c,0x00);
    rfm69_setreg(0x2d,0x00);

    //Sync Word abschalten
    rfm69_setregmask(0x2e,0x80,0x00);

    rfm69_set_rxtx_mode(2); // Standby
    rfm69_setreg(0x37,0x00); //Fixed Len // Seite 54 / 72

    ook_repeat_count = cmd->value[3];
    if ( ook_repeat_count < 4) { ook_repeat_count = 4; }
    if ( ook_repeat_count > 16) { ook_repeat_count = 16; }

    rfm_mode = 19;
    cmd->err = 0;

    // Alles weitere erledigt der main loop
    return;
#endif

#if RFM69_OOKMODE == 1

    uint32_t tx;
    uint8_t i,j;
    uint8_t bit;

    debug_printf("OOK Send: ");

    // Intertechno ITS-1500 zum anlernen muss 16 mal gesendet werden
    // zum schalten nur 2x

    for (j=0; j<4; j++)
    {
        tx = bits;
        ook_sendbit(2, 4); // sync senden
        for (i=0; i<32; i++)
        {
            tmp = tx & 0x80000000;
            if (tmp) {bit = 1;} else {bit = 0;}
            tx = (tx << 1);
            ook_sendbit(bit, 4);
            //debug_printf("%u,",bit);
        }
        ook_sendbit(2, 5); // sync senden
    }



    // TOOM Baumarkt Funksteckdose 1919384
/*
    for (j=0; j<4; j++)
    {
        ook_sendbit(2, 1); // sync senden
        tx = bits;
        for (i=0; i<24; i++)
        {
            tmp = tx & 0x80000000;
            if (tmp) {bit = 1;} else {bit = 0;}
            tx = (tx << 1);
            ook_sendbit(bit, 1);
            //debug_printf("%u,",bit);
        }
    }

*/
/*
    for (j=0; j<4; j++)
    {
        ook_sendbit(2, 2); // sync senden
        tx = bits;
        for (i=0; i<24; i++)
        {
            tmp = tx & 0x80000000;
            if (tmp) {bit = 1;} else {bit = 0;}
            tx = (tx << 1);
            ook_sendbit(bit, 2);
            //debug_printf("%u,",bit);
        }
    }

*/
    debug_printf("\r\n");

    // PB0 auf Eingang Pullup aus
    DDRB &=  ~_BV(PB0);
    PORTB &= ~_BV(PB0);

    //RFM69 in normalen modus schalten
    rfm69_setmode(0); // Packet mode
    rfm69_setmodulation(0); // 0=FSK 1=OOK
    rfm69_set_rxtx_mode(0); // RX Mode

    cmd->err = 0;
#endif
}

//https://github.com/Phunkafizer/RaspyRFM/blob/master/intertechno.py
// Transmit OOK via FIFO for Intertechno power socket
// https://www.sweetpi.de/blog/329/ein-ueberblick-ueber-433mhz-funksteckdosen-und-deren-protokolle
// Die Dokumente wiedersprechen sich
// https://www.mikrocontroller.net/topic/124084
// KORREKT:
// https://manual.pilight.org/protocols/433.92/switch/intertechno.html
// Bit Zeit = 280 micro Sekunden Verhltnis 1 zu 5 Kurz zu lang Sync = 1x Kurz High + 2xLang Low
void ook_send_message(void)
{
    rfm69_set_rxtx_mode(2); // Standby

    //FIFO fuellen
    uint8_t fifobytes[64] = { 0 };
    uint16_t bitpointer;
    uint32_t tx;
    uint32_t tmp;
    uint8_t i;

    bitpointer = 0;

    // Preamble
    fifobytes[0] = 0x80;
    fifobytes[1] = 0x00;
    bitpointer = 11;

    tx = ook_message;

    for (i=0; i<32; i++)
    //for (i=0; i<24; i++)
    {
        tmp = tx & 0x80000000;
        if (tmp)
        {
            //bit = 1;
            /* Intertechno */
            addbit(fifobytes,&bitpointer,1);
            addbit(fifobytes,&bitpointer,0);
            addbit(fifobytes,&bitpointer,0);
            addbit(fifobytes,&bitpointer,0);
            addbit(fifobytes,&bitpointer,0);
            addbit(fifobytes,&bitpointer,0);
            addbit(fifobytes,&bitpointer,1);
            addbit(fifobytes,&bitpointer,0);

            // TOOM
            /*
            addbit(fifobytes,&bitpointer,1);
            addbit(fifobytes,&bitpointer,1);
            addbit(fifobytes,&bitpointer,1);
            addbit(fifobytes,&bitpointer,0);
            */

        }
        else
        {
            //bit = 0;
            /* Intertechno */
            addbit(fifobytes,&bitpointer,1);
            addbit(fifobytes,&bitpointer,0);
            addbit(fifobytes,&bitpointer,1);
            addbit(fifobytes,&bitpointer,0);
            addbit(fifobytes,&bitpointer,0);
            addbit(fifobytes,&bitpointer,0);
            addbit(fifobytes,&bitpointer,0);
            addbit(fifobytes,&bitpointer,0);

            // TOOM
            /*
            addbit(fifobytes,&bitpointer,1);
            addbit(fifobytes,&bitpointer,0);
            addbit(fifobytes,&bitpointer,0);
            addbit(fifobytes,&bitpointer,0);
            */
        }
        tx = (tx << 1);
    }

    // footer pulse
    addbit(fifobytes,&bitpointer,1);
    // 10 millisekunden Pause
    bitpointer += 35;

    i = ( ( (bitpointer-1) / 8) +1);

    //debug_printf("BITS: %d BYTES: %d\r\n",bitpointer,i);
    rfm69_write_fifo(fifobytes,i+1);

    rfm69_setreg(0x37,0x00); //Fixed Len // Seite 54 / 72
    rfm69_setreg(0x38,i); // Payload Length = 0 // Seite 53


    // Fifo kontroll lesen
    /*
    uint8_t daten[64];
    rfm69_read_fifo(daten,20);
    for (i=0; i<20; i++)
    {
        debug_printf("FIFO %d val: 0x%02x\r\n",i,daten[i]);
    }
    */

    // TX Mode start
    rfm69_set_rxtx_mode(1); // TX Mode
    rfm_mode = 20;

    return;
}

void addbit(uint8_t *bitbuffer, uint16_t *position, uint8_t bit)
{
    uint8_t bytepos;
    uint8_t bitpos;
    uint8_t bitmask;

    bytepos = (*position) / 8;
    bitpos = (*position) - (bytepos * 8); // 0 bis 7
    bitpos = 7 - bitpos;
    bitmask = 0x01;
    bitmask = (1 << bitpos);

    //debug_printf("ADDBIT BYTE: %d BIT: %d VAL: %u MASK: %u\r\n",bytepos,bitpos,bit,bitmask);

    if (bit == 0)
    {
        // Keine Operation bit ist sowieso 0
        //bitbuffer[bytepos] &= ~bitmask;
    }
    else if (bit == 1)
    {
        bitbuffer[bytepos] |= bitmask;
    }

    (*position)++;
}

//Manuelles senden ueber den DIO2 data in pin
// ook mode = 1 heisst es wird manuelle gesendet ohne den FIFO zu benutzen
// #if RFM69_OOKMODE == 1
void ook_sendbit(uint8_t val, uint8_t protocol)
{
    /*
    uint16_t moduletime;
    moduletime = pgm_read_word(&ooktiming[protocol][0]);
    */
    // _delay_loop_2 = 4 CPU Cycles / count = 5 count = 1us @ 20MHz

    if (val == 0)
    {
        if (protocol == 1)
        {
            // 0-Bit
            PORTB |= _BV(PB0); // Port B0 auf 1
            _delay_loop_2(5*375); // 375 us an
            PORTB &= ~_BV(PB0); // Port B0 auf 0
            _delay_loop_2(5*1125); // 1145 us aus
        }
        else if (protocol == 2)
        {
            // 0-Bit
            PORTB |= _BV(PB0); // Port B0 auf 1
            _delay_loop_2(5*500); // 500 us an
            PORTB &= ~_BV(PB0); // Port B0 auf 0
            _delay_loop_2(5*1025); // 1025 us aus
        }
        else if (protocol == 3)
        {
            // 0-Bit
            PORTB |= _BV(PB0); // Port B0 auf 1
            _delay_loop_2(5*375); // 275 us an
            PORTB &= ~_BV(PB0); // Port B0 auf 0
            _delay_loop_2(5*1125); // 275 us aus
            PORTB |= _BV(PB0); // Port B0 auf 1
            _delay_loop_2(5*375); // 275 us an
            PORTB &= ~_BV(PB0); // Port B0 auf 0
            _delay_loop_2(5*1125); // 1225 us aus
        }
        else if (protocol == 4)
        {
            // 0-Bit Intertechno
            PORTB |= _BV(PB0); // Port B0 auf 1
            _delay_loop_2(5*275); // 275 us an
            PORTB &= ~_BV(PB0); // Port B0 auf 0
            _delay_loop_2(5*275); // 275 us aus
            PORTB |= _BV(PB0); // Port B0 auf 1
            _delay_loop_2(5*275); // 275 us an
            PORTB &= ~_BV(PB0); // Port B0 auf 0
            _delay_loop_2(5*1225); // 1225 us aus
        }
    }
    else if (val == 1)
    {
        if (protocol == 1)
        {
            // 1-Bit
            PORTB |= _BV(PB0); // Port B0 auf 1
            _delay_loop_2(5*1125); // 1117 us an
            PORTB &= ~_BV(PB0); // Port B0 auf 0
            _delay_loop_2(5*375); // 400 us aus
        }
        else if (protocol == 2)
        {
            // 1-Bit
            PORTB |= _BV(PB0); // Port B0 auf 1
            _delay_loop_2(5*1000); // 1000 us an
            PORTB &= ~_BV(PB0); // Port B0 auf 0
            _delay_loop_2(5*530); // 530 us aus
        }
        else if (protocol == 3)
        {
            // 1-Bit
            PORTB |= _BV(PB0); // Port B0 auf 1
            _delay_loop_2(5*1125); // 275 us an
            PORTB &= ~_BV(PB0); // Port B0 auf 0
            _delay_loop_2(5*375); // 1225 us aus
            PORTB |= _BV(PB0); // Port B0 auf 1
            _delay_loop_2(5*1125); // 275 us an
            PORTB &= ~_BV(PB0); // Port B0 auf 0
            _delay_loop_2(5*375); // 275 us aus
        }
        else if (protocol == 4)
        {
            // 1-Bit Intertechno
            PORTB |= _BV(PB0); // Port B0 auf 1
            _delay_loop_2(5*275); // 275 us an
            PORTB &= ~_BV(PB0); // Port B0 auf 0
            _delay_loop_2(5*1225); // 1225 us aus
            PORTB |= _BV(PB0); // Port B0 auf 1
            _delay_loop_2(5*275); // 275 us an
            PORTB &= ~_BV(PB0); // Port B0 auf 0
            _delay_loop_2(5*275); // 275 us aus
        }
    }
    else if (val == 2)
    {
        if (protocol == 1)
        {
            //SYNC
            PORTB |= _BV(PB0); // Port B0 auf 1
            _delay_loop_2(5*375); // 375 us an
            PORTB &= ~_BV(PB0); // Port B0 auf 0
            _delay_loop_2(5*2250); // 2280 us aus
//            _delay_loop_2(5*11250U); // 2280 us aus
        }
        else if (protocol == 2)
        {
            //SYNC
            PORTB |= _BV(PB0); // Port B0 auf 1
            _delay_loop_2(5*3000); // 3000 us an
            PORTB &= ~_BV(PB0); // Port B0 auf 0
            _delay_loop_2(5*7040U); // 7040 us aus
        }
        else if (protocol == 3)
        {
            //SYNC
            PORTB |= _BV(PB0); // Port B0 auf 1
            _delay_loop_2(5*375); // 275 us an
            PORTB &= ~_BV(PB0); // Port B0 auf 0
            _delay_loop_2(5*2625); // 2675 us aus
        }
        else if (protocol == 4)
        {
            //SYNC
            PORTB |= _BV(PB0); // Port B0 auf 1
            _delay_loop_2(5*275); // 275 us an
            PORTB &= ~_BV(PB0); // Port B0 auf 0
            //_delay_loop_2(5*2700); // 2700 us aus
            _delay_loop_2(5*2675); // 2675 us aus
        }
        else if (protocol == 5)
        {
            //SYNC
            PORTB |= _BV(PB0); // Port B0 auf 1
            _delay_loop_2(5*275); // 275 us an
            PORTB &= ~_BV(PB0); // Port B0 auf 0
            _delay_loop_2(5*10500U); // 10500 us aus
        }
    }
}
 //#endif

// Capture 433MHz OOK Codes
#if RFM69_OOK_CAPTURE == 1
static void cmd_ookcapture(struct cmd_data_t *cmd)
{
    if ( cmd->value[0] == 1)
    {
        pinlevel_pb0 = 0;
        ooktime_count = 0;

        // Pinchange interrupt fuer PacketMode AUSschalten
        PCMSK1 &= ~_BV(PCINT13); // On Pin PC5 = PCINT 13

        //RFM69 in den OOK RX Modus versetzen
        rfm69_setmode(3); // continous ohne sync
        rfm69_set_rxtx_mode(0);
        rfm69_setmodulation(1);

        // DIO0 an PB5 = RSSI
        rfm69_setreg(0x25,0x80);

        rfm_mode = 20;

        //min RSSI auf -70dBm setzen
        rfm69_setrssi(70);

        //OOKthresh
        rfm69_setreg(0x1b,0x43); // Seite 68 Peak
        //rfm69_setreg(0x1b,0x80); // Average
        rfm69_setreg(0x1d,6);

        // TIMER1 Starten
        // Timer1 fuer 433 MHz Funksteckdosen Steuerung
        // Normal Mode = Mode 0 Datenblatt Seite 136
        TCCR1A &= ~_BV(WGM10);
        TCCR1A &= ~_BV(WGM11);
        TCCR1B &= ~_BV(WGM12);
        TCCR1B &= ~_BV(WGM13);

        // Prescaler 64 = 3,2 uSec pro count bei 20 MHZ cputakt
        TCCR1B = _BV(CS10) | _BV(CS11); // Seite 139

        // Output compare macht interrupt nur einschalten wenn 433MHz gesendet werden soll
        //TIMSK1 |= _BV(OCIE1A); // Seite 136 + Input Caputre Register ICR1
        ti1_ovf_cnt = 0; // Timer 1 overflow count

        //TIMSK1 |= _BV(TOIE1); // Timer Overflow Interrup Enable

        //OCR1A = 255; // 255 * 12,8 uSec / Interrupt
        TCNT1 = 0;

        // Pinchange interrupt PB0 enable
        // Enable Pinchange Interrupt
        DDRB &= ~_BV(PB0); // PB0 is input
        PORTB &= ~_BV(PB0); // Pullup Resistor OFF
        PCICR |= _BV(PCIE0); // Enable Pin change Interrupt 0 Seite 74
        PCMSK0 |= _BV(PCINT0); // On Pin PB0 = PCINT0 Seite 75
    }

    if ( cmd->value[0] == 2)
    {
        rfm69_set_rxtx_mode(2); // standby

        //min RSSI auf -70dBm setzen
        rfm69_setrssi(70);

        // OOK
        rfm69_setmodulation(1);

        rfm69_setmode(0); // 0 = packet mode

        // DIO0 an PB5 = PayloadReady seite 69 und 48
        rfm69_setreg(0x25,0x40); // payload
        //rfm69_setreg(0x25,0xC0); // rssi

        //rfm69_setbaudrate(2666); // TOOM Baumarkt
        rfm69_setbaudrate(3333); // intertechno

        //Sync Word abschalten fifo wird gefullt ohne sync
        rfm69_setregmask(0x2e,0xC0,0x40);

        //OOK modulation rssi Threshold & decay parameters
        //rfm69_setreg(0x1b,0x80); // Average
        rfm69_setreg(0x1b,0x43); // Peak Seite 68
        rfm69_setreg(0x1d,70); // Peak Seite 68

        // fifo len = 32 byte // fixed len
        rfm69_setreg(0x37,0x08); // fifo conditions
        rfm69_setreg(0x38,50); // payload len

        // Fifo threshold seite 73
        rfm69_setreg(0x3c,45);

        // Enable Pinchange Interrupt
        DDRC &= ~_BV(PC5); // PC5 is input
        PORTC &= ~_BV(PC5); // Pullup Resistor OFF
        PCICR |= _BV(PCIE1); // Enable Pin change Interrupt 1
        PCMSK1 |= _BV(PCINT13); // On Pin PC5 = PCINT 13

        // RX Mode start
        //rfm69_setregmask(0x01,0x1C,0x10);
        rfm69_set_rxtx_mode(0);

        rfm_mode = 18;
    }

    if ( cmd->value[0] == 0)
    {
        rfm69_diable_capture();
    }
}

void rfm69_diable_capture(void)
{
    // Disable Pinchange Interrupt
    PCMSK0 &= ~_BV(PCINT0); // On Pin PB0 = PCINT0
    PCICR &= ~_BV(PCIE0); // Enable Pin change Interrupt 0

    // RFM69 Standby
    rfm69_set_rxtx_mode(2);

    serial_output((char*)"\r\n");

    TIMSK1 &= ~_BV(TOIE1); // Timer Overflow Interrupt Disable
    // Timer 1 anhalten
    TCCR1B = 0;

    //OOK Peak thresh
    rfm69_setreg(0x1b,0x40);
    rfm69_setreg(0x1c,6);

    rfm69_setmode(0);
    rfm69_setmodulation(0);
    rfm_mode = 0;
    // weitere settings fuer ook
    //min RSSI auf -92dBm setzen
    rfm69_setrssi(92);

    // Pinchange interrupt fuer PacketMode einschalten
    PCMSK1 |= _BV(PCINT13); // On Pin PC5 = PCINT 13

    //RFM69 in RX mode
    rfm69_set_rxtx_mode(0);
}
#endif // RFM69_OOK_CAPTURE
#endif


#if HELP_ENABLE == 1
void cmd_help(struct cmd_data_t *cmd)
{
    uint8_t cmd_id = 0;
    while (cli_cmd[cmd_id].parametercount > 0 && strlen_P(cli_cmd[cmd_id].help_text) > 0)
    {
        sprintf_P(uartbuf,cli_cmd[cmd_id].cmd_text);
        serial_output(uartbuf);

        sprintf_P(uartbuf,PSTR(": "));
        serial_output(uartbuf);

        sprintf_P(uartbuf,cli_cmd[cmd_id].help_text);
        serial_output(uartbuf);

        sprintf_P(uartbuf,PSTR("\r\n"));
        serial_output(uartbuf);
        cmd_id++;
    }
}
#endif

#if FNORD_NET_ENABLE == 1
void cmd_select(struct cmd_data_t *cmd)
{
    if (cmd->value[0] >= 0 && cmd->value[0] < 64)
    {
        // wenn die eigene fnordnet adresse selektiert wird
        // dann auf 0 setzen = local
        if (fnord_config.local_address == cmd->value[0] || cmd->value[0] == 0)
        {
            cmd->value[0] = 0;
            cmd->value[1] = 0;
            fnord_remote_address = 0;
            cmd->err = 1;
        }
        else
        {
            fnord_remote_address = cmd->value[0];
            cmd->value[1] = 0;

            fnord_transmit_setup((uint8_t)cmd->value[0]);
            cmd->value[1] = fnordnet_tx_status();
            if (cmd->value[1] == 1)
            {
                cmd->value[1] = 0;
                cmd->err = 1;
            }
            else
            {
                cmd->err = 3;
                fnord_remote_address = 0;
            }
        }
    }
    else
    {
        cmd->value[0] = 0;
        fnord_remote_address = 0;
        cmd->value[1] = 255;
        cmd->err = 5;
    }
}
#endif

static void cmd_defaults(struct cmd_data_t *cmd)
{
    set_cfg_defaults();

#if FNORD_NET_ENABLE == 1
    fnord_set_local_address(fnord_config.local_address);


#if RFM12_ENABLE == 1
        rfm12_setrssi(fnord_config.rssi);
        rfm12_setbaudrate(fnord_config.baudrate);
        rfm12_setchannel(fnord_config.channel);
        rfm12_settxpower(fnord_config.txpower);
        rfm12_init();
        debug_printf("RFM12: ADDR: %d RSSI: %d BAUD: %d CHAN %d TX: %d\r\n",fnord_config.local_address,fnord_config.rssi,fnord_config.baudrate,fnord_config.channel,fnord_config.txpower);
#endif

#if RFM69_ENABLE == 1
        rfm69_settxpower(fnord_config.txpower);
        rfm69_setchannel(fnord_config.channel);
        rfm69_setrssi(fnord_config.rssi);
        rfm69_setbaudrate(fnord_config.baudrate);
        rfm69_init();
        debug_printf("RFM69: ADDR: %d RSSI: %d BAUD: %d CHAN %d TX: %d\r\n",fnord_config.local_address,fnord_config.rssi,fnord_config.baudrate,fnord_config.channel,fnord_config.txpower);
#endif

#endif

    cmd->value[0] = fnord_config.rssi;
    cmd->value[1] = fnord_config.baudrate;
    cmd->value[2] = fnord_config.channel;
    cmd->value[3] = fnord_config.txpower;
}

static void cmd_savecfg(struct cmd_data_t *cmd)
{
    debug_printf("RFM: RSSI: %d BAUD: %d CHAN %d TX: %d\r\n",fnord_config.rssi,fnord_config.baudrate,fnord_config.channel,fnord_config.txpower);

    uint16_t status = eeprom_write_bytearray(fnord_config.raw,sizeof(fnord_config.raw),0);
    if ( status > 0)
    {
        cmd->err = 4;
        cmd->value[0] = status;
    }
    else
    {
        cmd->err = 1;
        cmd->value[0] = fnord_config.rssi;
        cmd->value[1] = fnord_config.baudrate;
        cmd->value[2] = fnord_config.channel;
        cmd->value[3] = fnord_config.txpower;
    }
}

/* nur zu debug zwecken */

//uint16_t xxprom_write_bytearray(uint8_t *bytes,uint16_t len, uint16_t offset)
//{
//    /* crc berechnen */
//    //uint16_t crc = crc_checksum(bytes,len);
//    uint16_t i;
//
//    len += offset;
//    for (i=offset; i<len; i++)
//    {
//        sprintf_P(uartbuf,PSTR("I: %d VAL: %u  "),i,bytes[(i-offset)]);
//        serial_output(uartbuf);
//    }
//    return 0;
//}


static void cmd_loadcfg(struct cmd_data_t *cmd)
{
    uint16_t status = eeprom_read_bytearray(fnord_config.raw,sizeof(fnord_config.raw),0);
    if (status != sizeof(fnord_config.raw))
    {
        cmd->err = 4;
        cmd->value[0] = status;
    }
    else
    {
#if RFM12_ENABLE == 1 || RFM69_ENABLE == 1
        cmd->err = 1;
        cmd_data.value[0] = (uint8_t)fnord_config.txpower;
        cmd_txpower(&cmd_data);
        cmd_data.value[0] = (uint8_t)fnord_config.channel;
        cmd_channel(&cmd_data);
        cmd_data.value[0] = (uint8_t)fnord_config.rssi;
        cmd_rssi(&cmd_data);
        cmd_data.value[0] = (uint16_t)fnord_config.baudrate;
        cmd_baudrate(&cmd_data);

#if RFM69_ENABLE == 1
        cmd_data.value[0] = (uint16_t)fnord_config.rfmzin;
        cmd_rfmzin(&cmd_data);
#endif
        cmd_data.value[0] = 0;

        cmd->err = 1;
        cmd->value[0] = fnord_config.rssi;
        cmd->value[1] = fnord_config.baudrate;
        cmd->value[2] = fnord_config.channel;
        cmd->value[3] = fnord_config.txpower;
#endif
    }
    debug_printf("RFM: ADDR: %d RSSI: %d BAUD: %d CHAN %d TX: %d\r\n",fnord_config.local_address,fnord_config.rssi,fnord_config.baudrate,fnord_config.channel,fnord_config.txpower);
}

static void cmd_json(struct cmd_data_t *cmd)
{
    if (cmd->value[0] == 0 || cmd->value[0] == 1 )
    {
        fnord_config.json = cmd->value[0];
        cmd->err = 0;
    }
    else
    {
        cmd->err = 6;
    }
}


static void cmd_out(struct cmd_data_t *cmd)
{
    uint8_t value1 = (uint8_t)cmd->value[0];
    uint8_t value2 = (uint8_t)cmd->value[1];

    //if (value1 >= 1 && value1 <=6 && value2 >=0 && value2 <=1) // value2 >=0 immer trur fr unsigned
    if (value1 >= 1 && value1 <=6 && value2 <=1)
    {

        DDRD |=  _BV(value1+1);

        if ( value2 == 1)
        {
            PORTD |= _BV(value1+1);
            cmd->value[2] = 0;
        }
        else if (value2 == 0)
        {
            PORTD &= ~_BV(value1+1);
            cmd->value[2] = 0;
        }
        cmd->err = 1;
    }
    else
    {
        cmd->err = 6;
    }
}

static void cmd_in(struct cmd_data_t *cmd)
{
    uint8_t value1 = cmd->value[0];
    uint8_t value2 = 0;
    if (value1 <= 6)
    {
        value1++;
        DDRD &=  ~_BV(value1);
        asm("nop");
        asm("nop");
        //PORTD |=  _BV(value1);
        value2 = (PIND & _BV(value1)) / _BV(value1);
        value2 = !value2;
        cmd->err = 1;
    }
    else if (value1 == 7 || value1 == 8)
    {
        value1 -= 7;
        DDRC &= ~_BV(value1);
        value2 = (PINC & _BV(value1)) / _BV(value1);
        value2 = !value2;
        value1 += 8;
        cmd->err = 1;
    }
    else
    {
        cmd->err = 6;
    }
    cmd->value[0] = (int16_t)value1-1;
    cmd->value[1] = (int16_t)value2;
}

void cmd_adclist(struct cmd_data_t *cmd)
{
    int16_t wert;
    int8_t cnt;
    if ( cmd->value[0] >= 0 && cmd->value[0] <= 5)
    {
        /* ADC enable */
        ADCSRA |= _BV(ADEN); // ADC Enable
        /* select channel */
        cmd->value[0]--;

        for (cnt=0; cnt<(uint8_t)cmd->value[1]; cnt++)
        {
            ADMUX &= 0b11110000;
            ADMUX |= (uint8_t)cmd->value[0];
            ADCSRA |= _BV(ADSC); // ADC Start Conversion
            /* Wait till complete */
            while ( (ADCSRA & _BV(ADSC)) == _BV(ADSC)) { asm("nop"); }
            /* fetch data */
            wert = ADCL + (ADCH<<8);
            out_printf("Cnt %02d Val %04d\r\n",cnt,wert);
            //for (uint8_t j=0;j < cmd->value[1];j++)
            //{
                _delay_loop_2(20000);
            //}
        }

        /* ADC disable */
        ADCSRA &= ~_BV(ADEN);
    }
    else
    {
        cmd->err = 6;
    }

}

void cmd_adcin(struct cmd_data_t *cmd)
{
    if ( cmd->value[0] >= 0 && cmd->value[0] <= 5)
    {
        uint16_t count;
        count = 0;
        cmd->value[0]--;
        /* ADC enable */
        ADCSRA |= _BV(ADEN); // ADC Enable
        /* select channel */
        ADMUX &= 0b11110000;
        ADMUX |= (uint8_t)cmd->value[0];
        ADCSRA |= _BV(ADSC); // ADC Start Conversion
        /* Wait till complete */
        while ( (ADCSRA & _BV(ADSC)) == _BV(ADSC)) {count++;}
        /* fetch data */
        cmd->value[1] = ADCL + (ADCH<<8);
        /* ADC disable */
        ADCSRA &= ~_BV(ADEN);
        cmd->value[0]++;
        cmd->err = 1;
    }
    else
    {
        cmd->err = 6;
    }
}

// Read an average on n values of one adc channel
void cmd_adcavgin(struct cmd_data_t *cmd)
{
    uint32_t avg_val = 0;
    uint16_t i;
    uint16_t divisor;

    if ( cmd->value[0] >= 0 && cmd->value[0] <= 5 && cmd->value[1] > 0 && cmd->value[1] < 10000)
    {
        divisor = (unsigned) cmd->value[1];

        cmd->value[0]--;
        /* ADC enable */
        ADCSRA |= _BV(ADEN); // ADC Enable
        /* select channel */
        ADMUX &= 0b11110000;
        ADMUX |= (uint8_t)cmd->value[0];

        for(i = 0 ; i < divisor ; i++ )
        {
            ADCSRA |= _BV(ADSC); // ADC Start Conversion
            /* Wait till complete */
            while ( (ADCSRA & _BV(ADSC)) == _BV(ADSC)) { asm("nop"); }
            /* fetch data */
            avg_val += ADCL + (ADCH<<8);
            _delay_loop_2(50); // wait 10 micro seconds
        }

        cmd->value[1] = avg_val / cmd->value[1];

        /* ADC disable */
        ADCSRA &= ~_BV(ADEN);
        cmd->value[0]++;
        cmd->err = 1;
    }
    else
    {
        cmd->err = 6;
    }
}

// Read an average on n values of one adc channel
void cmd_adcavginall(struct cmd_data_t *cmd)
{
    uint32_t avg_val[4];
    uint16_t i;
    uint16_t divisor;

    if ( cmd->value[0] > 0 && cmd->value[0] < 10000)
    {
        divisor = (unsigned) cmd->value[0];

        for (uint8_t count = 0; count < 4; count++ )
        {
            avg_val[count] = 0;
        }

        /* ADC enable */
        ADCSRA |= _BV(ADEN); // ADC Enable

        for(i = 0 ; i < divisor ; i++ )
        {
            for (uint8_t count = 0; count < 4; count++ )
            {
                /* select channel */
                ADMUX &= 0b11110000;
                ADMUX |= count;
                /* Just wait three cycles for mux to stabilize */
                asm("nop"); asm("nop"); asm("nop");
                ADCSRA |= _BV(ADSC); // ADC Start Conversion
                /* Wait till complete */
                while ( (ADCSRA & _BV(ADSC)) == _BV(ADSC)) { asm("nop"); }
                /* fetch data */
                avg_val[count] += ADCL + (ADCH<<8);
            }
            _delay_loop_2(50); // wait 10 micro seconds
        }


        for (uint8_t count = 0; count < 4; count++ )
        {
            debug_printf("D: %d VAL: %ld\r\n",divisor,avg_val[count]);
            cmd->value[count] = avg_val[count] / divisor;
        }

        /* ADC disable */
        ADCSRA &= ~_BV(ADEN);
        cmd->err = 1;
    }
    else
    {
        cmd->err = 6;
    }
}

void cmd_adcinall(struct cmd_data_t *cmd)
{
        uint8_t count;
        count = 0;

        for ( count = 0; count < 4; count++ )
        {
            /* ADC enable */
            ADCSRA |= _BV(ADEN); // ADC Enable
            /* select channel */
            ADMUX &= 0b11110000;
            ADMUX |= count;
            /* Just wait one cycle for mux to stabilize */
            asm("nop"); asm("nop"); asm("nop");
            ADCSRA |= _BV(ADSC); // ADC Start Conversion
            /* Wait till complete */
            while ( (ADCSRA & _BV(ADSC)) == _BV(ADSC)) { asm("nop"); }
            /* fetch data */
            cmd->value[count] = ADCL + (ADCH<<8);
            /* ADC disable */
            ADCSRA &= ~_BV(ADEN);
        }
        cmd->err = 1;
}

/* Set adc reference voltage */
void cmd_adcsetref(struct cmd_data_t *cmd)
{
    if ( cmd->value[0] == 1 || cmd->value[0] == 5)
    {
        /* Refernz Spannung = intern VCC */
        ADMUX |= _BV(REFS0);
        if ( cmd->value[0] == 1 )
        {
            /* Referenz 1.1 Volt internal reference ON*/
            ADMUX |= _BV(REFS1);
            fnord_config.adcref = 1;
        }
        else
        {
            /* Referenz 1.1 Volt internal reference OFF*/
            ADMUX &= ~_BV(REFS1);
            fnord_config.adcref = 5;
        }
    }
    else if ( cmd->value[0] == 0 )
    {
        if ((ADMUX & _BV(REFS1)) == 0)
        {
            cmd->value[0] = 5;
            cmd->value[1] = 5;
        }
        else
        {
            cmd->value[0] = 1;
            cmd->value[1] = 1;
        }
    }
    else
    {
        cmd->err = 6;
    }
}

void cmd_dacout(struct cmd_data_t *cmd)
{
    if ( cmd->value[0] == 1 || cmd->value[0] == 2)
    {
        uint8_t cmd_hi;
        uint8_t cmd_lo;
        uint8_t tmp;
        cmd_lo = (uint8_t)cmd->value[1];
        cmd_hi = (uint8_t)(cmd->value[1] >> 8);
        // Output channel select
        if (cmd->value[0] == 1)
        {   cmd_hi |= 0b00110000; }
        else
        {   cmd_hi |= 0b10110000; }
        DDRC |= _BV(PC4); // Ausgang
        PORTC &= ~_BV(PC4); // DAC aktiv
        // Daten ausgeben
        tmp = spi_transfer_byte(cmd_hi);
        tmp = spi_transfer_byte(cmd_lo);
        PORTC |= _BV(PC4); // DAC inaktiv
        cmd->err = 1;
    }
    else
    {
        cmd->err = 6;
    }
}

static void cmd_setmem(struct cmd_data_t *cmd)
{
    if (cmd->value[0] >= 0 && cmd->value[0] <=10)
    {
        merker.wert[cmd->value[0]] = cmd->value[1];
        cmd->err = 1;
    }
    else
    {
        cmd->err = 6;
    }
}

static void cmd_getmem(struct cmd_data_t *cmd)
{
    if (cmd->value[0] >= 0 && cmd->value[0] <=10)
    {
        cmd->value[1] = merker.wert[cmd->value[0]];
        cmd->err = 1;
    }
    else
    {
        cmd->err = 6;
    }
}

static void cmd_savemem(struct cmd_data_t *cmd)
{
    uint16_t status = eeprom_write_bytearray(merker.raw,sizeof(merker.raw),sizeof(fnord_config.raw)+2);
    if ( status > 0)
    {
        cmd->err = 4;
        cmd->value[0] = status;
    }
    else
    {
        cmd->err = 1;
    }
}

static void cmd_loadmem(struct cmd_data_t *cmd)
{
    uint16_t status = eeprom_read_bytearray(merker.raw,sizeof(merker.raw),sizeof(fnord_config.raw)+2);
    if (status != sizeof(merker.raw))
    {
        cmd->err = 4;
        cmd->value[0] = status;
    }
    else
    {
        cmd->err = 1;
    }
    uint8_t i;
    for(i=0; i <= 10;i++)
    {
        debug_printf("M: %u V: %d  ",i,merker.wert[i]);
    }
    debug_printf("\r\n");
}

#if EEPROM_DEBUG_CMDS == 1
static void cmd_rawread(struct cmd_data_t *cmd)
{
    uint8_t data[30];
    eeprom_read_raw_bytearray(data,20,0);
    cmd->err = 1;

    uint8_t i;
    for(i=0; i < 20;i++)
    {
        sprintf_P(uartbuf,PSTR("BYTE: %u V: %u  "),i,data[i]);
        serial_output(uartbuf);
    }
    sprintf_P(uartbuf,PSTR("\r\n"));
    serial_output(uartbuf);
}
#endif

#if RFM12_ENABLE == 1 || RFM69_ENABLE == 1
static void cmd_rfms(struct cmd_data_t *cmd)
{
#if RFM12_ENABLE == 1
    uint16_t rfs;
    rfs = rfm12_getstatus();
    cmd->value[0] = (int16_t)rfs;
    cmd->value[1] = (int16_t)rfm12_get_irqcount();
#endif
#if RFM69_ENABLE == 1
    cmd->value[0] = (int16_t)rfm69_get_chiprevision();
    cmd->value[1] = (int16_t)rfm69_get_irqcount();
#endif
}

#if RFM69_ENABLE == 1
static void cmd_rfmtemp(struct cmd_data_t *cmd)
{
    rfm69_set_rxtx_mode(2); // Standby mode
    rfm69_setreg(0x4e,0x08);
    while( rfm69_getreg(0x4e) > 0 )
    { asm("nop"); }

    cmd->value[0] = (int16_t)rfm69_getreg(0x4f);
    rfm69_set_rxtx_mode(0); // RX mode
}

static void cmd_rfmzin(struct cmd_data_t *cmd)
{
    if (cmd->value[0]  == 0 || cmd->value[0]  == 50 || cmd->value[0]  == 200)
    {
        if (cmd->value[0] == 0)
        {
            uint8_t val;
            val = rfm69_getreg(0x18);
            val = val & 0x80 ;
            if (val == 0)
            { cmd->value[0] = 50; }
            else
            { cmd->value[0] = 200; }
        }
        else
        {
            if (cmd->value[0] == 50)
            { rfm69_setregmask(0x18,0x80,0x00); }
            else
            { rfm69_setregmask(0x18,0x80,0x80); }

            fnord_config.rfmzin = cmd->value[0];
        }
        cmd->err = 1;
    }
}
#endif


static void cmd_channel(struct cmd_data_t *cmd)
{
#if RFM12_ENABLE == 1
    if ( cmd->value[0] > 0 && cmd->value[0] < 24)
    {
        rfm12_setchannel(cmd->value[0]);
        fnord_config.channel = (uint8_t)cmd->value[0];
        rfm12_init();
        cmd->err = 1;
    }
    else
    {
        cmd->err = 6;
    }
#endif
#if RFM69_ENABLE == 1
    if ( cmd->value[0] > 0 && cmd->value[0] < 24)
    {
        rfm69_setchannel(cmd->value[0]);
        fnord_config.channel = (uint8_t)cmd->value[0];
    }
    else
    {
        cmd->err = 6;
    }
#endif
}

#if RFM69_ENABLE == 1
static void cmd_frequency(struct cmd_data_t *cmd)
{
    uint32_t freq;
    freq = cmd->value[0] * 1000000L;
    freq += cmd->value[1] * 1000L;
    freq += cmd->value[2];
    //debug_printf("FREQ: %ld %d %d %d\r\n",freq,cmd->value[0],cmd->value[1],cmd->value[2]);
    rfm69_setfrequency(freq);
}
#endif


static void cmd_baudrate(struct cmd_data_t *cmd)
{
    fnord_config.baudrate = cmd->value[0];

#if RFM12_ENABLE == 1
    rfm12_setbaudrate( (uint16_t)cmd->value[0] );
    //    debug_printf("baudrate = %d",value1);
    rfm12_init();
#endif

#if RFM69_ENABLE == 1
    rfm69_setbaudrate( (uint16_t)cmd->value[0] );
#endif
}

static void cmd_txpower(struct cmd_data_t *cmd)
{
#if RFM12_ENABLE == 1
    if ( cmd->value[0] >= 0 && cmd->value[0] <= 7)
    {
        fnord_config.txpower = (uint8_t)cmd->value[0];
        cmd->value[0] = rfm12_settxpower( (uint8_t)cmd->value[0] );
        rfm12_init();
        cmd->err = 1;
    }
    else
    {
        cmd->err = 6;
    }
#endif
#if RFM69_ENABLE == 1
    if ( cmd->value[0] >= 0 && cmd->value[0] < 32)
    {
        fnord_config.txpower = (uint8_t)cmd->value[0];
        cmd->value[0] = rfm69_settxpower( (uint8_t)cmd->value[0] );
    }
    else
    {
        cmd->err = 6;
    }
#endif
}

static void cmd_rssi(struct cmd_data_t *cmd)
{
#if RFM12_ENABLE == 1
    if ( cmd->value[0] >= 0 && cmd->value[0] < 8)
    {
        rfm12_setrssi( (uint8_t)cmd->value[0] );
        fnord_config.rssi = (uint8_t)cmd->value[0];
        rfm12_init();
        cmd->err = 1;
    }
    else
    {
        cmd->err = 6;
    }
#endif
#if RFM69_ENABLE == 1
    if ( cmd->value[0] >= 0 && cmd->value[0] < 126)
    {
        rfm69_setrssi( (uint8_t)cmd->value[0] );
        fnord_config.rssi = (uint8_t)cmd->value[0];
        cmd->err = 1;
    }
    else
    {
        cmd->err = 6;
    }
#endif
}

static void cmd_rssi_scan(struct cmd_data_t *cmd)
{
#if RFM12_ENABLE == 1
#endif
#if RFM69_ENABLE == 1
    cmd->value[1] = rfm69_min_rssi(cmd->value[0],&cmd->value[2],&cmd->value[3]);
#endif
}

static void cmd_rssi_survey(struct cmd_data_t *cmd)
{
#if RFM12_ENABLE == 1
#endif
#if RFM69_ENABLE == 1
    uint8_t i;
    int16_t min,max,avg;
    cli();
    for (i=01; i<24; i++)
    {
        rfm69_setchannel(i);
        _delay_loop_2(500);
        avg = rfm69_min_rssi(cmd->value[0],&min,&max);
        out_printf("CH: %02d AVG: -%03d dBm MIN: -%03d dBm MAX: -%03d dBm\r\n",i,avg,min,max);
    }
    rfm69_setchannel(fnord_config.channel);
    sei();
#endif
}

static void cmd_setmod(struct cmd_data_t *cmd)
{
#if RFM69_ENABLE == 1
    rfm69_setmodulation((uint8_t)cmd->value[0]);
#endif
}


#endif



#if FNORD_NET_ENABLE == 1
static void cmd_sla(struct cmd_data_t *cmd)
{
    if ( cmd->value[0] > 0 && cmd->value[0] < 64)
    {
        fnord_set_local_address((uint8_t)cmd->value[0]);
        fnord_config.local_address = (uint8_t)cmd->value[0];
        srand(fnord_config.local_address);
        cmd->err = 1;
    }
    else
    {
        cmd->err = 5;
    }
}


static void cmd_setup(struct cmd_data_t *cmd)
{
    if ( cmd->value[0] > 0 && cmd->value[0] < 64)
    {
        fnord_transmit_setup((uint8_t)cmd->value[0]);
        cmd->value[1] = fnordnet_tx_status();
        // returns a 1 for ok but needs to be a 0 to indicate eveything ok
        if (cmd->value[1] == 1)
        { cmd->value[1] = 0;  cmd->err = 0;}
        else
        { cmd->err = 3; }

    }
    else
    {
        cmd->err = 1;
    }
}

// Ping a remote station, or local station just for fun
// returns pinged address parameter is arbitrary data not the node address to ping
// select node to ping with select cmd first
static void cmd_ping(struct cmd_data_t *cmd)
{
//    if ( cmd->value[0] >= -32768 && cmd->value[0] <= 32767)
//    {
//        cmd->value[1] = fnord_config.local_address;
//        cmd->err = 0;
//    }
//    else
//    {
//        cmd->err = 6;
//    }

        // cmd->value[0] ist int16 daher sind die werte begrenzt
        cmd->value[1] = fnord_config.local_address;
        cmd->err = 0;
}

static void cmd_status(struct cmd_data_t *cmd)
{
    cmd->value[0] = fnord_config.channel;
    cmd->value[1] = fnord_config.local_address;
    cmd->value[2] = fnord_config.txpower -18 ;
    cmd->value[3] = fnord_config.rssi;
    cmd->err = 0;
    //cmd->value[0] = fnordnet_tx_status();
}
#endif

/* RFM12 interrupt routine */

#if RFM12_ENABLE == 1
ISR(PCINT1_vect)
{
    rfm12_interrupt_routine();
}

// called when rfm12 received data
void rfm12_received_frame(uint8_t *data,uint8_t length)
{
    // What to do with a received frame?
    // send it to fnord-net protocol
    #if FNORD_NET_ENABLE == 1
    fnord_received_frame(data,length);
    #endif
}
#endif

#if RFM69_ENABLE == 1
ISR(PCINT1_vect)
{
    rfm69_interrupt_routine();
}

// called when rfm69 received data
void rfm69_received_frame(uint8_t *data,uint8_t length)
{
    // What to do with a received frame?
    // send it to fnord-net protocol
    #if FNORD_NET_ENABLE == 1
    fnord_received_frame(data,length);
    #endif
}
#endif



/* den Array zur auswertung der CLI und Fnordnet kommandos initialisiern */
void cli_cmd_init(void)
{
    uint8_t cnt;

    // for autosync with software using serial interface for example raspberry pi
    cnt=0;
    cli_cmd[cnt].cmd_text = PSTR("XXX");
    cli_cmd[cnt].return_text = PSTR("TESTING %d %d %d");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_test;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("test command");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("test");
    cli_cmd[cnt].return_text = PSTR("TESTING %d %d %d");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_test;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("test command");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("savecfg");
    cli_cmd[cnt].return_text = PSTR("Save config minRSSI:%d BAUD:%d CHAN:%d TXPWR:%d");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_savecfg;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Save config minRSSI BAUD CHAN TXPWR");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("loadcfg");
    cli_cmd[cnt].return_text = PSTR("Loading config minRSSI:%d BAUD:%d CHAN:%d TXPWR:%d");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_loadcfg;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Load config from EEPROM");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("savemem");
    cli_cmd[cnt].return_text = PSTR("Saving mem values to eeprom");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_savemem;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Save set/getmem parameters to EEPROM");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("loadmem");
    cli_cmd[cnt].return_text = PSTR("Loading mem values from eeprom");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_loadmem;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Load set/getmem parameters from EEPROM");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("out");
    cli_cmd[cnt].return_text = PSTR("Setze REL: %d Wert: %d");
    cli_cmd[cnt].parametercount = 3;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_out;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Set output [1-6] to value [0|1]");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("in");
    cli_cmd[cnt].return_text = PSTR("OC: %d Wert: %d");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_in;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Read input [1-8]");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("setmem");
    cli_cmd[cnt].return_text = PSTR("Setze mem %d auf Wert %d");
    cli_cmd[cnt].parametercount = 3;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_setmem;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Set memory [1-10] to value [+-32767]");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("getmem");
    cli_cmd[cnt].return_text = PSTR("Mem %d hat den Wert: %d");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_getmem;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Read memory [1-10]");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("adcinall");
    cli_cmd[cnt].return_text = PSTR("ADC Wert1: %d Wert2: %d Wert3: %d Wert4: %d");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_adcinall;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Read all ADC input channels");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("adcin");
    cli_cmd[cnt].return_text = PSTR("ADC Kanal: %d Wert: %d");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_adcin;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Read ADC [1-4] input");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("adcavginall");
    cli_cmd[cnt].return_text = PSTR("ADC AVG Wert1: %d Wert2: %d Wert3: %d Wert4: %d");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_adcavginall;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Read all ADC input average over [n] samples");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("adcavgin");
    cli_cmd[cnt].return_text = PSTR("ADC Kanal: %d duchschittswert: %d");
    cli_cmd[cnt].parametercount = 3;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_adcavgin;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Read ADC [1-4] input average over [n] samples");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("adclist");
    cli_cmd[cnt].return_text = PSTR("OK");
    cli_cmd[cnt].parametercount = 3;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_adclist;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Show adc [1-4] value [n] times");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("adcsetref");
    cli_cmd[cnt].return_text = PSTR("ADC Referenz %d Volt");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_adcsetref;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Set ADC Reference Voltage [0(read)|1(set)|5(set)] Volts");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("dacout");
    cli_cmd[cnt].return_text = PSTR("DAC Kanal: %d Wert: %d");
    cli_cmd[cnt].parametercount = 3;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_dacout;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Set DAC [1|2] output to value [0-4095]");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("defaults");
    cli_cmd[cnt].return_text = PSTR("RFM: RSSI: %d BAUD: %d CHAN %d TX: %d");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_defaults;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Set RFM CFG to defaults");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("version");
    cli_cmd[cnt].return_text = PSTR("Version: %d.%d");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_version;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Show firmware version");
#endif

#if DHT11_ENABLE == 1
    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("temp");
    cli_cmd[cnt].return_text = PSTR("Temperature: %d Humidity: %d");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_temp;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Get temperature*10 and humidity*10 from DHT22 Module");
#endif
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("memfree");
    cli_cmd[cnt].return_text = PSTR("Mem free %d bytes");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_memfree;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Get amount of free RAM");
#endif

#if FNORD_NET_ENABLE == 1
    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("sla");
    cli_cmd[cnt].return_text = PSTR("set local address to %d");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_sla;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Set fnordnet local address");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("setup");
    cli_cmd[cnt].return_text = PSTR("send setup frame to: %d Status: %d");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_setup;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Send fnordnet setup frame to address");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("select");
    cli_cmd[cnt].return_text = PSTR("Selected Target %d");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_select;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Select module for remote cmd execution [fnordnet-address 1-63]");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("status");
    cli_cmd[cnt].return_text = PSTR("RFM Status: Chan:%d Addr:%d TXpwr:%d dBm rssi:-%d dBm");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_status;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Get RFM status");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("ping");
    cli_cmd[cnt].return_text = PSTR("Traget: you send %d i am node %d");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_ping;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("ping [data] return data send as parameter by currently selected node");
#endif
#endif // endif FNORDNET ENABLE


#if RFM12_ENABLE == 1 || RFM69_ENABLE == 1

#if RFM12_ENABLE == 1
    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("rfms");
    cli_cmd[cnt].return_text = PSTR("rfmstatus: 0x%04x IRQ: Count %d");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_rfms;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Get RFM Module Status");
#endif
#endif

#if RFM69_ENABLE == 1
    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("rfms");
    cli_cmd[cnt].return_text = PSTR("RFM69 Chip Revision: %d IRQ Count: %d");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_rfms;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Get RFM Module Status");
#endif
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("channel");
    cli_cmd[cnt].return_text = PSTR("setting channel = %d");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_channel;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Set RFM Module channel [1-23]");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("baudrate");
    cli_cmd[cnt].return_text = PSTR("baudrate = %d BPS");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_baudrate;
#if HELP_ENABLE == 1
#if RFM12_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Set RFM Module baudrate [0-7]");
#endif
#if RFM69_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Set RFM Module baudrate [bps]");
#endif
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("txpower");
    cli_cmd[cnt].return_text = PSTR("txpower = %d dBm");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_txpower;
#if HELP_ENABLE == 1
#if RFM12_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Set RFM Module txpower [0-7] 7=min 0=max");
#endif
#if RFM69_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Set RFM Module [txpower] 0=min 31=max");
#endif
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("rssiscan");
    cli_cmd[cnt].return_text = PSTR("scanned %d times avg rssi = -%d dBm min -%d dBm max -%d dBm");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_rssi_scan;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Scan for noise (min RSSI)");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("rssisurvey");
    cli_cmd[cnt].return_text = PSTR("Finished");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_rssi_survey;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Scan all channels for noise (min RSSI) [n] times");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("rssi");
    cli_cmd[cnt].return_text = PSTR("min rssi = -%d dbm");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_rssi;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Set RFM Module min rssi for receiver to start");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("setmod");
    cli_cmd[cnt].return_text = PSTR("modulation: %d");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_setmod;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Set Modulation 0=FSK 1=OOK, dont use just an experiment");
#endif

#if RFM69_ENABLE == 1
    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("freq");
    cli_cmd[cnt].return_text = PSTR("setting frequency to %d MHz + %d KHz + %d Hz");
    cli_cmd[cnt].parametercount = 4;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_frequency;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Set RFM69 Module frequency [x]*1000000 + [y]*1000 + [z] Hz");
#endif
#endif
#endif //endif RFM12ENABLE || RFM69ENABLE



#if EEPROM_DEBUG_CMDS == 1
    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("rawread");
    cli_cmd[cnt].return_text = PSTR("Rawread %d bytes");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_rawread;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("");
#endif
#endif

#if RFM69_DEBUG_CMDS == 1 && RFM69_ENABLE == 1
    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("rfm");
    cli_cmd[cnt].return_text = PSTR("rfm69cw RET: 0x%04x");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_rfm;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("rfm69cw read test");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("rfm");
    cli_cmd[cnt].return_text = PSTR("rfm69cw write REG: %d DATA: %d");
    cli_cmd[cnt].parametercount = 3;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_rfm_write;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("rfm69cw write test");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("fifow");
    cli_cmd[cnt].return_text = PSTR("rfm69cw write fifo");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_fifow;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("rfm69cw write fifo");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("fifor");
    cli_cmd[cnt].return_text = PSTR("rfm69cw read fifo");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_fifor;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("rfm69cw read fifo");
#endif

#endif // #if RFM69_DEBUG_CMDS == 1 && RFM69_ENABLE == 1


#if HELP_ENABLE == 1
    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("json");
    cli_cmd[cnt].return_text = PSTR("JSON is %u");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_json;
    cli_cmd[cnt].help_text = PSTR("Turn JSON Output ON=1/OFF=0");
#endif

#if HELP_ENABLE == 1
    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("help");
    cli_cmd[cnt].return_text = PSTR("Help for all commands");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_help;
    cli_cmd[cnt].help_text = PSTR("This list");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("itsset");
    cli_cmd[cnt].return_text = PSTR("Set its Address to %d %d");
    cli_cmd[cnt].parametercount = 3;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_itsset;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Set Intertechno Address [Adr 16Bit] [Adr 10Bit MSB]");
#endif

// http://www.fhemwiki.de/wiki/Intertechno_Code_Berechnung#Selbstlernende_Intertechno_Funksteckdosen_.28z.B._ITR-1500.29
// Example command
// its 0xaaaa 0x5555 0x11 0x05 turn on 5 repetitions
// its 0xaaaa 0x5555 0x10 0x05 turn off
// for initialisation of a new power socket use 16 repetitions !!
#if RFM69_OOK == 1
    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("its");
    cli_cmd[cnt].return_text = PSTR("Sending its command");
    cli_cmd[cnt].parametercount = 5;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_ooksend;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Send Intertechno Cmd [Adr 16Bit] [Adr 10Bit MSB] 0x00[group,on-off] 0x00[repeat]");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("toom");
    cli_cmd[cnt].return_text = PSTR("Sending toom command");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_toom;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Send Toom 1919384 Cmd [Adr 10Bit] [on=1 off=0]");
#endif

//https://wiki.pilight.org/doku.php/arctech_switch
#if RFM69_OOK_CAPTURE == 1
    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("capture");
    cli_cmd[cnt].return_text = PSTR("OOK Capture enabled");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_ookcapture;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Analyze signals from 433MHz remote control");
#endif
#endif // RFM69_OOK_CAPTURE
#endif // RFM_69OOK

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("bootloader");
    cli_cmd[cnt].return_text = PSTR("");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_bootloader;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Start bootloader, press Enter TWICE!!");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("crc32");
    cli_cmd[cnt].return_text = PSTR("");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 1;
    cli_cmd[cnt].func = &cmd_crc32;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("crc32 test");
#endif

#if RFM69_ENABLE == 1

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("rfmtemp");
    cli_cmd[cnt].return_text = PSTR("RFM69 Temperature: %d");
    cli_cmd[cnt].parametercount = 1;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_rfmtemp;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Get RFM module temperature from buildin sensor");
#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("rfmzin");
    cli_cmd[cnt].return_text = PSTR("RFM69 Antenna Zin: %d Ohm");
    cli_cmd[cnt].parametercount = 2;
    cli_cmd[cnt].always_local = 0;
    cli_cmd[cnt].func = &cmd_rfmzin;
#if HELP_ENABLE == 1
    cli_cmd[cnt].help_text = PSTR("Get/Set RFM69 module antenna base Z (0|50|200)");
#endif

#endif

    cnt++;
    cli_cmd[cnt].cmd_text = PSTR("");
    cli_cmd[cnt].return_text = PSTR("");
    cli_cmd[cnt].parametercount = 0;
    cli_cmd[cnt].always_local = 0;
    //cli_cmd[cnt].func = &cmd_;
    debug_printf("Registered %u commands\r\n",cnt);
}

void cli_err_init(void)
{
    errormessage[0] = PSTR("");
    errormessage[1] = PSTR("OK");
    errormessage[2] = PSTR("ERROR");
    errormessage[3] = PSTR("ERROR: Remote site does not answer");
    errormessage[4] = PSTR("ERROR: EEPROM write error on byte %d");
    errormessage[5] = PSTR("ERROR: Invalid fnordnet address 1-63");
    errormessage[6] = PSTR("ERROR: Invalid values");
    errormessage[7] = PSTR("ERROR: parity mismatch");
    errormessage[8] = PSTR("ERROR: Requeted device not found");
    errormessage[9] = PSTR("");
    errormessage[10] = PSTR("");
    errormessage[11] = PSTR("");
    errormessage[12] = PSTR("");
    errormessage[13] = PSTR("");
    errormessage[14] = PSTR("cmd returned unknown error please fix!");
}
