#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <stdio.h>

#include "rfm12.h"
#include "fnord-net.h"
#include "jackalope.h"

#if RFM12_ENABLE == 1

static uint8_t rfm12_rx(void);
static void rfm12_tx_finish(void);
static void rfm12_wait_tx_ready(void);
static uint16_t rfm12_sendcmd(uint16_t);
static void rfm12_settransmitter(uint8_t,uint8_t);
static void rfm12_setreceiver(uint8_t,uint8_t,uint8_t,uint8_t);
static void rfm12_setdatafilter(uint8_t);
static void rfm12_rx_start(void);
static uint8_t rfm12_get_mode(void);

/* Gobale Variablen RFM12 */
static uint32_t volatile rfm12_irqcnt;

static uint8_t rfm12_rx_data[RFM12_RX_BUFLEN];
static uint8_t rfm12_rx_pointer;

static uint8_t rfm12_tx_data[RFM12_TX_BUFLEN];
static uint8_t rfm12_tx_pointer;
static uint8_t rfm12_tx_len;

static uint8_t volatile rfm12_gotirq;
volatile uint8_t rfm_mode; // Status are we in rx=0 or tx=1 mode

/* RFM12 Parameters */
// Channels with 75 KHz Bandwidth
const uint16_t PROGMEM rfm12_channels[23] = {1238,1268,1298,1328,1358,1388,1418,1448,1478,1508,1538,1568,1598,1628,1658,1688,1718,1748,1778,1808,1838,1868,1898};

uint8_t channel = 11;
uint16_t f0;

// Baudrate approx 9600 BPS
uint16_t baudrate = 9578;

// Baudrate apporox 1200 BPS
//uint16_t baudrate = 1197;

// -21 db tx power
uint8_t txpower = 7;

// +- 30 KHZ tx frequenzy shift
uint8_t txbw = 1;

// +- 75 KHZ RX frequence bandwidth
uint8_t rxbw = 6;
uint8_t rssi = 1;
uint8_t dqd = 3;
uint8_t gain = 0;



/*
        rfm12_setbaudrate(9578); // 9578 BPS or 1197 BPS to get a nearly even divisor
        rfm12_setfrequency(1568); // f0 = 433.92 MHz
        rfm12_settransmitter(7,7); // (CMD Nr:10)  7 = -21dB , 7 = 7+1*15kHz FSK Frequency Shift = 120kHz

        // Configure Receiver (5. Receiver control command)
        // Valid data indicator = slow = 3, Bandwidth =5=134 MHz, LNA Gain=0=-0dB, RSSI =5=-73dBm
        rfm12_setreceiver(3,5,0,5);
*/

/* macro for calculating frequency value out of frequency in MHz */
#define RFM12FREQ(freq) ((freq-430.0)/0.0025)


/* --------------------- RFM-12 Functions ----------------------- */


void rfm12_init(void)
{
        rfm12_sendcmd(0xC0E0);      /* 14. Low bat & mcu clock -> 10MHz */
        rfm12_sendcmd(0x80D7);      /* 1. conf set cmd, el=1 ef=1 b1=0 b0=1 x3..x0=7 crystal load=12pF*/
        rfm12_sendcmd(0xC2AB);      /* 6. data filter -> al=1 ml=0 s=0 f2=0 f1=1 f0=1 -> DQD=3*/
        rfm12_sendcmd(0xCA81);      /* 7. fifo and reset cmd, 8bit->IRQ al=0 ff=0 dr=1 */
        rfm12_sendcmd(0xE000);      /* 12. wakeup timer cmd, all values 0 = wakeup disable */
        rfm12_sendcmd(0xC800);      /* 13. low duty cycle cmd, all values 0 = low duty cycle disable */
        rfm12_sendcmd(0xC4F7);      /* 9 AFC cmd, do not change default value */
        rfm12_sendcmd(0x82D9);      /* 8. Pwr mgmt*/
        rfm12_sendcmd(0x0000);      /* read status */

        /*
        //RFM12 Parameter
        uint16_t f0 = 1568;
        uint16_t baudrate = 9578;
        uint8_t txpower = 7;
        uint8_t txbw = 1;
        uint8_t rxbw = 6;
        uint8_t rssi = 1;
        uint8_t dqd = 3;
        */

        f0 = pgm_read_word(&rfm12_channels[channel]);

        // Set Baudrate
        rfm12_setbaudrate(baudrate); // 9578 BPS or 1197 BPS to get a nearly even divisor
        rfm12_setfrequency(f0); // f0 = 1568 = 433.92 MHz
        rfm12_settransmitter(txpower,txbw); // (CMD Nr:10)  7 = -21dB , 1 = 1+1*15kHz FSK Frequency Shift = 30kHz

        // Configure Receiver (5. Receiver control command)
        // Valid data indicator = slow = 3, Bandwidth =5=134 MHz, LNA Gain=0=-0dB, RSSI =5=-73dBm
        rfm12_setreceiver(2,rxbw,gain,rssi);
        rfm12_setdatafilter(dqd); // 0-7 dqd level for valid data detection

        // Enable receiver
        rfm12_rx_start();

        // Enable Interrupt
        DDRC &= ~_BV(PC5); // PC5 is input
        PORTC &= ~_BV(PC5); // Pullup Resistor OFF
        PCICR |= _BV(PCIE1); // Enable Pin change Interrupt 1
        PCMSK1 |= _BV(PCINT13); // On Pin PC5 = PCINT 13

        // Deselect RFM12 chip
        DDRB |= _BV(PB1);
        PORTB |= _BV(PB1);

        // Für was ist das??
        //DDRD |= _BV(PD2);
        //PORTD &= ~_BV(PD2);

}

// Check if there is a RFM12 onboard
uint8_t rfm12_check_presence(void)
{
    return 0;
}

uint16_t rfm12_getstatus(void)
{
    return rfm12_sendcmd(0x0000);
}

uint16_t rfm12_sendcmd(uint16_t cmd)
{
    uint16_t retval;
    uint8_t lo;
    uint8_t hi;

    uint8_t cmd_lo;
    uint8_t cmd_hi;

    cmd_lo = (uint8_t)cmd;
    cmd_hi = (uint8_t)(cmd >> 8);


    // RFM12 Chip Select
    PORTB &= ~_BV(PB1);

    hi = spi_transfer_byte(cmd_hi);
    lo = spi_transfer_byte(cmd_lo);

    // RFM12 Chip DeSelect
    PORTB |= _BV(PB1);


    retval = (hi << 8) + lo;
    return retval;
}

/* RFM12 4. Data Rate Command */
/* Values for nearly even divisors 38312, 28734, 19156, 9578, 4789, 2535, 2395, 1197 */
void rfm12_setbaudrate(uint16_t baud)
{
    baudrate = baud;
    if (baud<663)
        return;
    if (baud<5400)  // Baudrate= 344827,58621/(R+1)/(1+CS*7)
        rfm12_sendcmd(0xC680|((43104/baud)-1));
    else
        rfm12_sendcmd(0xC600|((344828UL/baud)-1));
}

/* RFM12 Command Number 3. Frequency setting Command
 * freq = ((frequency-430.0)/0.0025)
 * 433,05 MHz - 434,79 MHz
 */
void rfm12_setfrequency(uint16_t freq)
{
    f0 = freq;
    if (freq<96)    // min. 430,2400MHz
        freq=96;
    else if (freq>3903) // max. 439,7575MHz
        freq=3903;
    rfm12_sendcmd(0xA000|freq);
}

/* 10. TX Configuration Control Command
 * mod = Bandwidth = (mod + 1) * 15 kHz
 * power = tx power = 0=0db, 1=-3dB, 2=-6dB, ...3db Steps ... 7 = -21dB
 */
void rfm12_settransmitter(uint8_t power,uint8_t mod)
{
    txpower = power ;
    rfm12_sendcmd(0x9800|(power&7)|((mod&15)<<4));
}

/* 5. Receiver control Command
 * Receiver setup
 */
void rfm12_setreceiver(uint8_t vdi,uint8_t bandwidth, uint8_t lnagain, uint8_t drssi)
{
    rfm12_sendcmd(0x9400 |((vdi & 3) << 8)|((bandwidth & 7) << 5)|((lnagain & 3) << 3) | (drssi & 7));
}

/* 6. Datafilter command
 * Paramter = DQD Level
 */
void rfm12_setdatafilter(uint8_t dataquality)
{
    rfm12_sendcmd(0xC2A8 | (dataquality & 7));
}

// Set RX/TX Channel. Value 1 to 23
void rfm12_setchannel(uint8_t newchannel)
{
    if (newchannel > 0 && newchannel < 24)
    {
        channel = --newchannel;
        f0 = pgm_read_word(&rfm12_channels[channel]);
        rfm12_setfrequency(f0);
    }
}

// power = tx power = 0=0db, 1=-3dB, 2=-6dB, ...3db Steps ... 7=-21dB
int16_t rfm12_settxpower(uint8_t newtxpower)
{
    int16_t retval = 0;
    if (newtxpower >= 0 && newtxpower <8)
    {
        txpower = newtxpower;
        rfm12_settransmitter(txpower,txbw);
        retval = newtxpower * -3;
    }
    return retval;
}

// Empfindlichkeit des Empfängers in db
// Wenn die Umgebung sehr verrauscht ist kann der Empfänger
// etwas unempfindlicher gemacht werden
//rssidb = {-103,-97,-91,-85,-79,-73,-67,-61};
void rfm12_setrssi(uint8_t newrssi)
{
    if ( newrssi >= 0 && newrssi < 8)
    {
        rssi = newrssi;
        rfm12_setreceiver(2,rxbw,gain,rssi);
    }
}

void rfm12_tx_bytestream(uint8_t *bytestream,uint8_t bytecount)
{

    rfm12_tx_data[0] = 0xAA;
    rfm12_tx_data[1] = 0x2D;
    rfm12_tx_data[2] = 0xD4;
    rfm12_tx_data[3] = bytecount+1; // incl. null byte das als letztes uebertragen wird RFM69 Seite 52 ?? Unklar

    for (uint8_t i = 0 ; i< bytecount; i++)
    {
        rfm12_tx_data[i+4] = bytestream[i];
    }

    rfm12_tx_data[bytecount + 4] = 0x55;
    rfm12_tx_len = bytecount + 5;

    rfm12_tx_pointer = 0;

    /* 2. Power Mgmt cmd, et=1=PLL on, es=1=synthesizer on, ex=1 crystal osc on */
    /* receiver = off ->  er=off and ebb=off */

    // load Registers with preamble
    rfm12_sendcmd(0xB8AA);
    rfm12_sendcmd(0xB8AA);


    // Startup Transmitter
    rfm_mode = 1;


    // the first interrupt will fire before the command is finished
    // so turn it off for command execution in order not to mess things up
    cli();
    rfm12_sendcmd(0x8239);
    sei();


    /* The interrupt routine is now taking care
     * of all data transmission
     */
//-------------------------------------------------

    // Wait for PLL to sync
    //_delay_loop_2(500);

/*
    // Wait for first preamble byte to be transmitted
    rfm12_wait_tx_ready();

    rfm12_sendcmd(0xB8AA);
    rfm12_wait_tx_ready();

    // Transmit Magic preamble
    rfm12_sendcmd(0xB82D);
    rfm12_wait_tx_ready();

    rfm12_sendcmd(0xB8D4);
    rfm12_wait_tx_ready();


    for (uint8_t i=0 ; i < 10 ; i++ )
    {
        rfm12_sendcmd(0xB800|data);
        rfm12_wait_tx_ready();
      data++;
    }
*/
    // Dummy byte as last byte
    //rfm12_sendcmd(0xB800);
    //rfm12_wait_tx_ready();

    //rfm12_tx_finish();

}

void rfm12_tx_finish(void)
{
    /* 2. Power Mgmt cmd, et=0=PLL off, es=0=synthesizer off, ex=1 crystal osc on */
    rfm12_sendcmd(0x8208);
    rfm12_rx_start();
#if FNORD_NET_ENABLE == 1
    set_fnord_tx_in_progress(0);
#endif
}

void rfm12_wait_tx_ready(void)
{
    while(rfm12_gotirq == 0){}
    rfm12_gotirq = 0 ;
}

/* Start RFM12 Receiver circuit */
void rfm12_rx_start(void)
{
    /* 2. Power Mgmt cmd, er=1=receiver on, ebb=1=baseband unit on, ex=1 crystal osc on */
    rfm12_sendcmd(0x82D9);         // RX on

    /* 7. FIFO and Reset mode Command */
    rfm12_sendcmd(0xCA81);         // enable FIFO with 8 bit buffer stop any ongoing reception

    /* 7. FIFO and Reset mode Command */
    rfm12_sendcmd(0xCA83);         // ff=1 enable FIFO after synchron pattern reception

    rfm_mode = 0;
}

/* fetch a byte form receive fifo */
uint8_t rfm12_rx(void)
{
    uint8_t data;
    data=(uint8_t)rfm12_sendcmd(0xB000);
    return data;
}

uint32_t rfm12_get_irqcount(void)
{
    return rfm12_irqcnt;
}

uint8_t rfm12_get_mode(void)
{
    return rfm_mode;
}


/* this function scans for the noise level */
int8_t rfm12_min_rssi(uint16_t testcycles)
{
    int8_t rssidb[8] = {-103,-97,-91,-85,-79,-73,-67,-61};
    uint16_t j = 0;
    uint8_t err;
    uint8_t errmax = testcycles / 100;

    for(uint8_t i = 0 ; i < 8 ; i++)
    {
        rfm12_setreceiver(2,rxbw,gain,i);
        err = 0;

        // wait 100us
        _delay_loop_2(500);

        for (j=0 ; j < testcycles ; j++)
        {
            _delay_loop_2(500);
            if ( (rfm12_getstatus() & 0x0100) == 0 )
            { }
            else
            { err++; }

            if (err >= errmax)
            {
                break;
            }

        }

        if (err < errmax)
        {
            rfm12_setreceiver(2,rxbw,gain,rssi);
            return rssidb[i];
        }
    }
    rfm12_setreceiver(2,rxbw,gain,rssi);
    return j;
}

#if RSSI_SCAN == 1
void rfm12_rssi_scan(uint16_t testcycles)
{
    uint8_t chan;
    int8_t noise_db;
    char msg[20];
    for (chan = 0 ; chan < 23 ; chan ++)
    {
        rfm12_setfrequency(pgm_read_word(&rfm12_channels[chan]));
        noise_db = rfm12_min_rssi(testcycles);
        ansi_locate(chan+5,1);
        sprintf_P(msg,PSTR("C:%d %d"),chan,noise_db);
        debug_output(msg);
    }
    rfm12_setfrequency(f0);
}
#endif

void rfm12_interrupt_routine(void)
{
    // Falling level only
    if ( (PINC & 0b00100000) == 32)
    { return; }

    // RX Mode
    if (rfm_mode == 0)
    {
        uint16_t rfm12_status;
        rfm12_status =  rfm12_getstatus();

        // FIFO is not empty, read one byte form FIFO
        if ( rfm12_status & 0x8000 )
        {

#if FNORD_NET_ENABLE == 1

            rfm12_rx_data[rfm12_rx_pointer] = (uint8_t)rfm12_sendcmd(0xB000);
            rfm12_rx_pointer++;

// TODO als erstes byte wird die frame laenge uebertragen
//            if (rfm12_rx_pointer == RFM12_RX_BUFLEN)
            if (rfm12_rx_pointer == (rfm12_rx_data[0]+1) || rfm12_rx_pointer == RFM12_RX_BUFLEN)
            {
                rfm12_received_frame(rfm12_rx_data,rfm12_rx_pointer);
                rfm12_rx_pointer = 0;

                // STOP receiving by reenableing the receiver -> bug in RFM12
                rfm12_rx_start();
            }

            fnord_clear_silence_count();
#else
            //rfm12_data[rfm12_rx_pointer] = rfm12_rx();
            rfm12_rx_data[rfm12_rx_pointer] = (uint8_t)rfm12_sendcmd(0xB000);
            rfm12_rx_pointer++;

            if (rfm12_rx_pointer == RFM12_RX_BUFLEN)
            {
                // Stop receiving (garbage) data
                rfm12_rx_pointer = RFM12_RX_BUFLEN;
                rfm12_rx_start();
            }

#endif
        }

    }


    // TX Mode
    if (rfm_mode == 1 && (rfm12_getstatus() & 0x8000) )
    {
        if (rfm12_tx_pointer < rfm12_tx_len)
        {
            rfm12_sendcmd(0xB800|rfm12_tx_data[rfm12_tx_pointer]);
            rfm12_tx_pointer++;

#if FNORD_NET_ENABLE == 1
            fnord_clear_silence_count();
#endif
        }
        else
        {
            rfm12_tx_pointer=0;
            rfm12_tx_len=0;
            rfm12_tx_finish(); // start receive mode and set rfm_mode
        }
    }

    rfm12_irqcnt++;
}

#endif
